﻿/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "ChannelBridge.h"
#include "../../DatafoxLibraryIV/libraryversion.h"
#include "../../DatafoxLibraryIV/DatafoxLibraryIV.h"
#include "../Common/eeprom.h"
#include "../Common/crc.h"
#include <stdlib.h>
#include <time.h>
#ifdef _MSC_VER
#include <io.h>
#else
#include <sys/uio.h>
#endif
#include "../File/filestudioxml.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

ChannelBridge::ChannelBridge( short channelId ) : CDFLog( channelId )
{
    // Kanalnummer setzen
    setChannelId( channelId );

    // Protokollzustaende der angesprochenen Geraete
    memset( _protocolVersion, 0, sizeof(_protocolVersion) );

    // Bevorzugt AEIII+ pruefen
    _preferAEIII = false;

    // Protocolversion für späteren Pufferzugriff halten.
    _protocolVersionRecordDescription = 4;
    _protocolVersionListDescription = 4;

    // Ist die bevorzugte Version der Dockingstation == 1, dann wird diese Klasse aktiviert.
    _channel4.setTimeboyStationEnabled( _dockingPreferType == 1 );

    // Datensatzpuffer loeschen
    _recordSize = 0;
    _recordError  = 0;
    memset( _recordData, 0, sizeof(_recordData) );

    // Wiederholungsversuche bei checkDevice
    _devicePollRetry = 0;

    // Trace aus Log aktivieren
    setTraceFileName( m_szFileNameTrace );

    // Info zum binden des Kanals an ein bestimmtes Geraet.
    memset( &_bindInfo, 0, sizeof(_bindInfo) );
    _bindInfo.typeDevice = -1; // 0 wuerde PZE-MasterIV sein. -1 ist kein Geraet.

    _channel4.setTrace( m_traceMode );

#ifdef _MSC_VER
    // Initialize the critical section one time only.
    InitializeCriticalSection( &m_mutexAPI ); 
    _threadIdOpen = 0;
#else
    pthread_mutex_init( &m_mutexAPI, NULL );
#endif
}

ChannelBridge::~ChannelBridge()
{
#ifdef _MSC_VER
    // Release resources used by the critical section object.
    DeleteCriticalSection( &m_mutexAPI );
#else
    pthread_mutex_destroy( &m_mutexAPI );
#endif
}

int ChannelBridge::tryEnterCriticalSection()
{
#ifdef _MSC_VER
    return TryEnterCriticalSection( &m_mutexAPI );
#else
    return !pthread_mutex_trylock( &m_mutexAPI );
#endif
}

void ChannelBridge::enterCriticalSection()
{
#ifdef _MSC_VER
    EnterCriticalSection( &m_mutexAPI );
#else
    pthread_mutex_lock( &m_mutexAPI );
#endif
}

void ChannelBridge::leaveCriticalSection()
{
#ifdef _MSC_VER
    LeaveCriticalSection( &m_mutexAPI );
#else
    pthread_mutex_unlock( &m_mutexAPI );
#endif
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
int ChannelBridge::openAEIII( int iType, char *szComString, int iComValue, int iComTimeOut )
{
    // Bevorzugt AEIII+ pruefen
    _preferAEIII = true;

    return open( IV_DEFAULT_DEVICEID, iType, szComString, iComValue, iComTimeOut );
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
int ChannelBridge::open( int deviceID, int iType, const std::string &szComString, int iComValue, int iComTimeOut )
{
    int res = IV_ERROR_FAIL;

    // Kanal darf nicht gebunden sein.
    if ( isBind() ) 
        return IV_ERROR_FAIL;

    // Kanal 4 ist Massgebend
    if ( _channel4.getType() == Channel4Base::Undefined ) 
    {
        if ( _channel4._commString == szComString ) 
        {
            // Bestimmt die Gueltigkeit der Protokollversion

            // Alle Teilnehmer erhalten unbekanntes Protokoll
            memset(_protocolVersion, 0, sizeof(_protocolVersion));
            // Ggf. unbekannte Dockingstation
            _channel4.resetTimeboyStation();
        }

        // Je nach Typ der Verbindung die Klasse initialisieren.
        switch (iType) 
        {
        case 1: // RS232 direkt (ohne autobaud)
            // Schnittstelle oeffnen
            _channel4.setAutobaud(false);
            res = _channel4.open(Channel4Base::Serial, szComString, iComValue, iComTimeOut);
            break;

        case 3: // TCP/IP
            // Schnittstelle oeffnen
            _channel4.setAutobaud(false);
            res = _channel4.open(Channel4Base::Socket, szComString, iComValue, iComTimeOut);
            break;

        case 4: // RS232 direkt mit autobaud
            // Schnittstelle oeffnen
            _channel4.setAutobaud(true);
            res = _channel4.open(Channel4Base::Serial, szComString, iComValue, iComTimeOut);
            break;

        default: // Unbekannter Schnittstellentyp
            res = IV_ERROR_FAIL;
        }

        // Nach Erfolg die Grundeinstellung wiederherstellen
        if (res == IV_ERROR_SUCCESS) 
        {
            setDeviceId( deviceID );

            if ( _channel4._commString == szComString ) 
            {
                // Bestimmt die Gueltigkeit der Protokollversion

                // Alle Teilnehmer erhalten unbekanntes Protokoll
                memset( _protocolVersion, 0, sizeof(_protocolVersion) );
                // Ggf. unbekannte Dockingstation
                _channel4.resetTimeboyStation();
            }
        }
    } 

    return res;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
void ChannelBridge::close(void)
{
    // Kanal 4 ist Massgebend
    _channel4.close();

    // Bevorzugt AEIII+ pruefen
    _preferAEIII = false;
}

// ##################################################################
/**
* @brief Ermitteln ob die Schnittstelle geoeffnet ist.
*/
int ChannelBridge::isOpen()
{
    // Channel4 ist Massgebend!
    return _channel4.isOpen();
}

// ##################################################################
/**
* @brief Schnittstelle nach Fehler bereinigen.
*/
void ChannelBridge::resetWithError( int err ) 
{
    // Gehaltene Protokollversion auswerten
    switch( _protocolVersion[(_channel4.getDeviceIdRemote() == 0xFE) ? 0 : _channel4.getDeviceIdRemote()] ) 
    {
    case 0:
    case 1:
    case 2:
    case 3: // Anstehende Daten verwerfen
        if ( err == IV_ERROR_CRC_PACKET || err == IV_ERROR_CHANNEL_COMMUNICATION || err == IV_ERROR_RECEIVE ) 
        {
            _channel4.flush();
        }
    case 4: 
        // Protocolversion zuruecksetzen
        if ( err == IV_ERROR_CHANNEL_COMMUNICATION || err == IV_ERROR_RECEIVE ) 
        {
            setProtocolVersion( getDeviceIdRemote(), 0 );
            // Ggf. unbekannte Dockingstation
            _channel4.resetTimeboyStation();
        }
    }
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
short ChannelBridge::setDeviceIdRemote(short deviceIDRemote)
{
    short deviceIDPrev = _channel4.getDeviceIdRemote();

    _channel1_2.SetDeviceIDRemote((unsigned char)deviceIDRemote);
    _channel4.setDeviceIdRemote(deviceIDRemote);

    return deviceIDPrev;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
short ChannelBridge::getDeviceIdRemote()
{
    return _channel4.getDeviceIdRemote();
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
short ChannelBridge::setDeviceId(short deviceID)
{
    short deviceIDPrev = _channel4.getDeviceId();

    _channel1_2.SetDeviceID((unsigned char)deviceID);
    _channel4.setDeviceId(deviceID);

    return deviceIDPrev;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
short ChannelBridge::getDeviceId()
{
    return _channel4.getDeviceId();
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
int ChannelBridge::comReleaseHandle()
{ 
    // CChannel4
    return _channel4.detachChannel();
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
int ChannelBridge::comUseHandle( HANDLE hCom )
{
   // CChannel4
   return  _channel4.attachChannel( Channel4Base::Serial, hCom, 800 );
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
HANDLE ChannelBridge::getComPort()
{
    HANDLE hCom = INVALID_HANDLE_VALUE;

    if ( _channel4.getType() == Channel4Base::Serial )
    {
        hCom = _channel4._comm;
    }
    else if ( _channel4.getType() == Channel4Base::Socket )
    {
        hCom = (HANDLE)_channel4._socket;
    }

    return hCom;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
int ChannelBridge::setComPort(HANDLE hCom, int iType, int iComValue, int iTimeOut)
{
    switch (iType)
    {
    case 0:
    case 2:
        // Muss undefiniert sein oder mit gewuenschtem uebereinstimmen
        if ((_channel4.getType() == Channel4Base::Undefined) || 
            ((_channel4.getType() == Channel4Base::Serial) && (_channel4._comm == hCom)) )
        {
            if (_channel4.changeStateSerial(hCom, iComValue, iTimeOut) != IV_ERROR_SUCCESS)
            {
                return IV_ERROR_FAIL;
            }

            // CChannel4
            _channel4.attachChannel(Channel4Base::Serial, hCom, iTimeOut);
        }
        else
        {
            return IV_ERROR_FAIL;
        }
        break;

    case 1:
    case 3:
        // Muss undefiniert sein oder mit gewuenschtem uebereinstimmen
        if ((_channel4.getType() == Channel4Base::Undefined) || 
            ((_channel4.getType() == Channel4Base::Socket) && (_channel4._socket == (SOCKET)hCom)) )
        {
            // CChannel4
            _channel4.attachChannel(Channel4Base::Socket, hCom, iTimeOut);
        }
        else
        {
            return IV_ERROR_FAIL;
        }
        break;
    }

    return IV_ERROR_SUCCESS;
}

void ChannelBridge::setCallBackFunction(int (*pFunction)(void))
{
    _channel1_2.SetCBFunction(pFunction);
    _channel4.setCallBackFunction(pFunction);
}

// Setzen des Logdateinamen.
void ChannelBridge::setLogFileName(const std::string &szLogFileName)
{
    _channel4.setTraceFilePath(szLogFileName);
}	


//#########################################################
//####################### HELPER ##########################
//#########################################################

int ChannelBridge::read(void *buffer, int receiveCount, int *receivedCount)
{
    return _channel4.readData(buffer, receiveCount, (int&)*receivedCount);
}

int ChannelBridge::write(void *buffer, int transmitCount, int *transmittedCount)
{
    return _channel4.writeData(buffer, transmitCount, (int&)*transmittedCount);
}

// Aufbereiten und senden eines Protokollpakets
int ChannelBridge::writePacket(short deviceIDRemote, unsigned char type, short module, short func, unsigned char *buf, int count, int transmitTrials)
{
    return _channel4.writePacket(deviceIDRemote, type, module, func, buf, count, transmitTrials);
}

// Empfang und aufbereiten eines Protokollpakets
int ChannelBridge::readPacket(short deviceID, unsigned char &type, short &module, short &func, unsigned char *buf, int &count, int receiveTrials)
{
    return _channel4.readPacket(deviceID, type, module, func, buf, count, receiveTrials);
}


// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
unsigned char ChannelBridge::getProtocolVersion(short deviceIDRemote)
{
    switch(_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote]) 
    {
    case 1: // Version 1 des Protokolls
    case 2: // Version 2 des MM-Protokolls
    case 3: // Mehrfachdockingstation
    case 4: // Version 4 des MM-Protokolls
        return _protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote];

    default: // Ermitteln der Protokollversion

        // Bevorzugt auf AEIII+ pruefen?
        if (_preferAEIII && (_channel1_2.Request((unsigned char)deviceIDRemote) == TRUE)) 
        { 
            // Pruefen ob Protokollversion 1 vorhanden
            return _protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = 1;	
        } 
        else if ( isOpen() && _channel4.isTimeboySlot(deviceIDRemote) ) 
        { 
            // Pruefen ob Protokollversion 4 vorhanden
            return _protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = 4;
        } 
        else if ( isOpen() && (_channel4.ping(deviceIDRemote, 0, 0) >= IV_ERROR_SUCCESS || _channel4.getLastResult() == -IV_ERROR_NEED_USER_PASSWORD) ) 
        { 
            // Pruefen ob Protokollversion 4 vorhanden
            return _protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = 4;
        } 
        else if ( isOpen() ) // Schnittstelle kann ggf. aufgrund von Fehlern bereits geschlossen sein.
        {	
            // Handelt es sich ggf. um eine Dockingstation der Version 1
            if ( _dockingPreferType == 2 ) 
            { 
                _channel4.setTimeboyStationEnabled( true );
                if ( _channel4.isTimeboySlot(deviceIDRemote) )
                    // Pruefen ob Protokollversion 4 vorhanden
                    return _protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = 4;

                _channel4.setTimeboyStationEnabled( false );
            }

            // Hoher Modbustimeout aelterer Firmwarestaende uebergehen.
            Sleep(100);
            if (_channel1_2.Request2((unsigned char)deviceIDRemote) == TRUE) 
            { 
                // Pruefen ob Protokollversion 2 vorhanden
                return _protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = 2;
            } 
            else if ( isOpen() ) // Schnittstelle kann ggf. aufgrund von Fehlern bereits geschlossen sein.
            {
                // Hoher Modbustimeout aelterer Firmwarestaende uebergehen.
                Sleep(100);
                if (_channel1_2.Request((unsigned char)deviceIDRemote) == TRUE) 
                { 
                    // Pruefen ob Protokollversion 1 vorhanden
                    return _protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = 1;
                }
            }
        }
    }

    return 0;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
unsigned char ChannelBridge::setProtocolVersion(short deviceIDRemote, unsigned char version)
{
    unsigned char old = _protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote];

    _protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = version;

    return old;
}

//////////////////////////////////////////////////////////////////////
// Members
//////////////////////////////////////////////////////////////////////

int ChannelBridge::collectData(short deviceIDRemote, int *piError)
{
    int iRes = -1;
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    // Loeschen des Datenpuffer
    if (_recordData[0])
    {
        _recordSize = 0;
        _recordError = 0;
        memset( _recordData, 0, sizeof(_recordData) );
    }

    switch ( getProtocolVersion(getDeviceIdRemote()) )
    {
    case 1:
    case 2:
        // Datensatz ermitteln
        iRes = _channel1_2.ReadRecord((unsigned char)getDeviceIdRemote(), _recordData, &_recordSize, &_recordError);
        break;

    case 4: {
        unsigned char status = 0;
        _recordSize = 0;
        _recordData[0] = (unsigned char)getDeviceIdRemote();

        // Befehlsausfuehrung starten.
        iRes = _channel4.getRecord(getDeviceIdRemote(), &_recordData[3], _recordSize, status);

        iRes = resultToError(iRes, &_recordError);
        if (iRes)
        {
            // OK
            iRes = status;

            // Groesze anpassen
            _recordSize += 5;
        }
        else
        {
            iRes = -1;
        }
        break;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
        return -1;
    }

    switch ( iRes )
    {
    case 0: // Keine Datensaetze vorhanden		
        _recordSize = 0;
        _recordError = 0;
        memset( _recordData, 0, sizeof(_recordData) );
        return 0;

    case 1: // Online Datensatz
    case 2: // Wiederholter Online Datensatz
        return 2;

    case 3: // Offline Datensatz
    case 4: // Wiederholter Offline Datensatz
        return 1;

    default:
        // Fehler bei Ermittlung
        *piError = _recordError;
    }

    return -1;
}

int ChannelBridge::getDatensatz(unsigned char *pucBuf, int *piLength, int *piError)
{
    int iRes = 0;

    // Wenn BusNum des letzten Datensatzpakets 0 ist dann sind 
    // keine gueltigen Daten vorhanden.
    if (_recordData[0] == 0)
    {
        return 0;
    }

    // Daten in Puffer uebergeben
    //unsigned char ucTmp[MAX_MB_PAKET_DATA * 3 + 1];
    int i;
    // Zu grosze Datensaetze abfangen!
    if ((_recordSize - 6) > MAX_MB_PAKET_DATA)
    {
        _recordSize -= (_recordSize - 6) - MAX_MB_PAKET_DATA;
    }
    for (i=0; i < (_recordSize - 6); i++)
    {
        pucBuf[i] = _recordData[i + 4];
        //sprintf((char*)&ucTmp[i * 3], "%02X ", (unsigned char)_recordData[i + 4]);
    }
    //ucTmp[i * 3] = 0;

    *piLength = _recordSize - 6;

    //theApp.m_log.Append("\t%s\t%i", ucTmp, *piLength);

    switch(getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:
        // Datensatz quittieren (3x -> Versuchen)
        for (i=0; (i < 3) && (iRes = _channel1_2.QuitRecord((unsigned char)getDeviceIdRemote(), _recordData,
            &_recordSize, &_recordError)) == -1; i++);

            // Quittierung erfolgreich?
            if (iRes)
            {	
                // Naechsten Datensatz mit Duplikattest anfordern
                for (i=0; (i < 3) && (iRes = _channel1_2.ReadRecord((unsigned char)getDeviceIdRemote(), _recordData,
                    &_recordSize, &_recordError)); i++);
            } 
            break;

    case 4: {
        unsigned char status = 0;

        // Befehlsausfuehrung starten.
        iRes = _channel4.setRecord(getDeviceIdRemote(), status);

        iRes = resultToError(iRes, &_recordError);
        if (iRes)
        {
            // OK
            // naechsten Datensatz lesen.
            unsigned char status = 0;
            _recordSize = 0;
            _recordData[0] = (unsigned char)getDeviceIdRemote();

            // Befehlsausfuehrung starten.
            iRes = _channel4.getRecord(getDeviceIdRemote(), &_recordData[3], _recordSize, status);

            iRes = resultToError(iRes, &_recordError);
            if (iRes)
            {
                // OK
                iRes = status;

                // Groesze anpassen (wegen veralgemeinerter Ausfuehrung am Funktionsanfang)
                _recordSize += 5;
            }
            else
            {
                iRes = -1;
            }
        }
        else
        {
            iRes = -1;
        }
        break;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
        return -1;
    }

    switch ( iRes )
    {
    case 0: // Keine Datensaetzte vorhanden
        // Aktuell uebergebener ist letzter moeglicher, bei Quittierungsmisserfolg.
        _recordSize = 0;
        _recordError = 0;
        memset(_recordData, 0, sizeof(_recordData));
        return 2;

        // Online Datensaetze
    case 1: // Online Datensatz
    case 2: // Wiederholter Online Datensatz
        return 1;

        // Offline Datensaetze
    case 3: // Offline Datensatz
    case 4: // Wiederholter Offline Datensatz
        return 1;

    default:
        *piError = _recordError;
    }

    return -1;
}

int ChannelBridge::systemReadRecord(int deviceIDRemote, char *buf, int *length, int *error)
{
    int iRes = -1;
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch(getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2: 
        *error = IV_ERROR_VERSION_FIRMWARE;
        return -1;

    case 4: {
        // OK
        // naechsten Datensatz lesen.
        unsigned char status = 0;
        _recordSize = 0;
        _recordData[0] = (unsigned char)getDeviceIdRemote();

        // Befehlsausfuehrung starten.
        iRes = _channel4.systemReadRecord(getDeviceIdRemote(), &_recordData[3], _recordSize, status);

        iRes = resultToError(iRes, &_recordError);
        if (iRes)
        {
            // OK
            iRes = status;

            // Groesse anpassen
            _recordSize += 5;
        }
        else
        {
            iRes = -1;
        }
        break;}

    default:
        *error = IV_ERROR_UNREACHABLE;
        return -1;
    }

    //	TRACE("( %i )\n", iRes);
    switch ( iRes )
    {
    case 0: // Keine Datensaetze vorhanden
        _recordSize = 0;
        _recordError = 0;
        memset(_recordData, 0, sizeof(_recordData));
        return 0;

    case 1: // System Datensatz
        // Daten in Puffer uebergeben, zu grosse Datensaetze sind abfangen!
        for (int i=0; i < (_recordSize - 6); i++)
        {
            buf[i] = _recordData[i + 4];
        }
        *length = _recordSize - 6;
        return iRes;

    default:
        // Fehler bei Ermittlung
        *error = _recordError;
        resetWithError(*error);
        break;
    }

    return -1;
}

int ChannelBridge::systemQuitRecord(int deviceIDRemote, int *error)
{
    int iRes = -1;
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2: 
        *error = IV_ERROR_VERSION_FIRMWARE;
        return -1;

    case 4: {
        // OK
        // naechsten Datensatz lesen.
        unsigned char status = 0;

        // Befehlsausfuehrung starten.
        iRes = _channel4.systemQuitRecord(getDeviceIdRemote(), status);

        iRes = resultToError(iRes, &_recordError);
        if (iRes)
        {
            // OK
            iRes = status;
        }
        else
        {
            iRes = -1;
        }
        break;}

    default:
        *error = IV_ERROR_UNREACHABLE;
        return -1;
    }

    //	TRACE("( %i )\n", iRes);
    switch ( iRes )
    {
    case 0: // Keine Datensaetze vorhanden
        return 0;

    case 1: // Datensatz quitiert
        return 1;

    default:
        // Fehlercode uebernehmen
        *error = _recordError;
    }

    return -1;
}

int ChannelBridge::systemRestoreRecords(short deviceIDRemote, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2: 
        *piError = IV_ERROR_VERSION_FIRMWARE;
        return 0;

    case 4: { // PZEMasterIV mit Protokollversion 4
        int iRes;

        // Befehlsausfuehrung starten.
        iRes = _channel4.systemRestoreRecords(getDeviceIdRemote());

        return resultToError(iRes, piError);}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::readRecord(int deviceIDRemote, char *buf, int *length, int *error)
{
    int iRes = -1;
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:

        // Datensatz ermitteln
        iRes = _channel1_2.ReadRecord((unsigned char)getDeviceIdRemote(), _recordData, &_recordSize, &_recordError);
        break;

    case 4: {
        // OK
        // naechsten Datensatz lesen.
        unsigned char status = 0;
        _recordSize = 0;
        _recordData[0] = (unsigned char)getDeviceIdRemote();

        // Befehlsausfuehrung starten.
        iRes = _channel4.getRecord(getDeviceIdRemote(), &_recordData[3], _recordSize, status);

        iRes = resultToError(iRes, &_recordError);
        if (iRes)
        {
            // OK
            iRes = status;

            // Groesse anpassen
            _recordSize += 5;
        }
        else
        {
            iRes = -1;
        }
        break;}

    default:
        *error = IV_ERROR_UNREACHABLE;
        return -1;
    }

    //	TRACE("( %i )\n", iRes);
    switch ( iRes )
    {
    case 0: // Keine Datensaetze vorhanden
        _recordSize = 0;
        _recordError = 0;
        memset(_recordData, 0, sizeof(_recordData));
        return 0;

    case 1: // Online Datensatz
    case 2: // Wiederholter Online Datensatz
    case 3: // Offline Datensatz
    case 4: // Wiederholter Offline Datensatz

        // Daten in Puffer uebergeben
        int i;
        // Zu grosse Datensaetze abfangen!
        if ((_recordSize - 6) > MAX_MB_PAKET_DATA)
        {
            _recordSize -= (_recordSize - 6) - MAX_MB_PAKET_DATA;
        }
        for (i=0; i < (_recordSize - 6); i++)
        {
            buf[i] = _recordData[i + 4];
        }

        *length = _recordSize - 6;
        return iRes;

    default:
        // Fehler bei Ermittlung
        *error = _recordError;
        resetWithError(*error);
        break;
    }

    return -1;
}

int ChannelBridge::quitRecord(int deviceIDRemote, int *error)
{
    int iRes = -1;
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:

        // Datensatz ermitteln
        iRes = _channel1_2.QuitRecord((unsigned char)getDeviceIdRemote(), _recordData, &_recordSize, &_recordError);
        break;

    case 4: {
        // OK
        // naechsten Datensatz lesen.
        unsigned char status = 0;

        // Befehlsausfuehrung starten.
        iRes = _channel4.setRecord(getDeviceIdRemote(), status);

        iRes = resultToError(iRes, &_recordError);
        if (iRes)
        {
            // OK
            iRes = status;
        }
        else
        {
            iRes = -1;
        }
        break;}

    default:
        *error = IV_ERROR_UNREACHABLE;
        return -1;
    }

    //	TRACE("( %i )\n", iRes);
    switch ( iRes ) 
    {
    case 0: // Keine Quittung noetig	
        return 0;

    case 1: // Online Datensatzquit.
    case 2: // Wiederholte Online Datensatzquit.
    case 3: // Offline Datensatzquit.
    case 4: // Wiederholte Offline Datensatzquit.
        return iRes;

    default:
        // Fehler bei Ermittlung
        *error = _recordError;
        resetWithError(*error);
        break;
    }

    return -1;
}

// Loescht alle Datensaetze
int ChannelBridge::clearData(short deviceIDRemote, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote())) 
    {
    case 1:
    case 2:
        return _channel1_2.ClearData((unsigned char)getDeviceIdRemote(), piError);

    case 4: { // PZEMasterIV mit Protokollversion 4
        int iRes;

        // Befehlsausfuehrung starten.
        iRes = _channel4.clearRecords(getDeviceIdRemote());

        return resultToError(iRes, piError);}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Stellt den Datensatzzeiger an Ringbeginn. Alle Datensaetze wieder abrufbar.
int ChannelBridge::restoreRecords(short deviceIDRemote, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote())) {
    case 1:
    case 2: 
        *piError = IV_ERROR_VERSION_FIRMWARE;
        return 0;

    case 4: { // PZEMasterIV mit Protokollversion 4
        int iRes;

        // Befehlsausfuehrung starten.
        iRes = _channel4.restoreRecords(getDeviceIdRemote());

        return resultToError(iRes, piError);}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::getDateTime(short deviceIDRemote, unsigned char *pucDateTime, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:

        if(!_channel1_2.ReadMem(MB_GET_TIME, 7, pucDateTime, piError)) 
        {		
            return 0;
        }
        return 1;

    case 4: {
        int iRes = IV_ERROR_SUCCESS;
        char szTime[20];

        // Befehlsausfuehrung starten.
        iRes = _channel4.getRTC(getDeviceIdRemote(), szTime);

        iRes = resultToError(iRes, piError);
        if (iRes)
        {
            // OK
            pucDateTime[0] = (szTime[0] - 0x30) * 10 + szTime[1] - 0x30;
            pucDateTime[1] = (szTime[2] - 0x30) * 10 + szTime[3] - 0x30;

            pucDateTime[2] = (szTime[5] - 0x30) * 10 + szTime[6] - 0x30;

            pucDateTime[3] = (szTime[8] - 0x30) * 10 + szTime[9] - 0x30;

            pucDateTime[4] = (szTime[11] - 0x30) * 10 + szTime[12] - 0x30;
            pucDateTime[5] = (szTime[14] - 0x30) * 10 + szTime[15] - 0x30;
            pucDateTime[6] = (szTime[17] - 0x30) * 10 + szTime[18] - 0x30;
        }

        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::setDateTime(short deviceIDRemote, unsigned char *pucDateTime, int *piError)
{
    unsigned char ucDateTime[8];

    if (pucDateTime == NULL ||
        (pucDateTime[0] != 20) ||
        (pucDateTime[1] > 99) ||
        (pucDateTime[2] < 1) || (pucDateTime[2] > 12) || 
        (pucDateTime[3] < 1) || (pucDateTime[3] > 31) ||
        (pucDateTime[4] > 23) || (pucDateTime[5] > 59) || (pucDateTime[6] > 59))
    {	
        time_t tm_time;
        struct tm *ptm;
        tm_time = time(NULL);
        ptm = localtime(&tm_time);

        ucDateTime[0] = (ptm->tm_year + 1900) / 100;
        ucDateTime[1] = (ptm->tm_year + 1900) % 100;
        ucDateTime[2] = ptm->tm_mon + 1;
        ucDateTime[3] = ptm->tm_mday;
        ucDateTime[4] = ptm->tm_hour;
        ucDateTime[5] = ptm->tm_min;
        ucDateTime[6] = ptm->tm_sec;
    }
    else
    {
        memcpy(ucDateTime, pucDateTime, 7);
    }

    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    int iRes = IV_ERROR_SUCCESS;
    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:

        if(!_channel1_2.WriteMem(MB_SET_TIME, 7, ucDateTime, piError))
        {	
            return 0;
        }
        return 1;

    case 4: {
        char szTime[20];

        sprintf(szTime, "%02i%02i-%02i-%02i %02i:%02i:%02i", 
            ucDateTime[0], ucDateTime[1], ucDateTime[2], ucDateTime[3], 
            ucDateTime[4], ucDateTime[5], ucDateTime[6]);

        // Befehlsausfuehrung starten.
        iRes = _channel4.setRTC(getDeviceIdRemote(), szTime);

        return resultToError(iRes, piError);}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::sendInfotext(short deviceIDRemote, char *szMessage, int iLength, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1: {
        unsigned char Buffer[245];
        int counter;

        Buffer[0] = 0xFF;
        Buffer[1] = 0;
        Buffer[2] = 0;

        for(counter=0; counter < iLength; counter++)
        {
            Buffer[counter + 3] = szMessage[counter];
        }

        if(!_channel1_2.WriteMem(INFOTEXT_MEMORY, iLength + 3, &Buffer[0], piError))
        {			
            return 0;
        }

        return 1;}

    case 2:
        *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
        return 0;

    case 4: {
        int iRes = IV_ERROR_SUCCESS;

        // Befehlsausfuehrung starten.
        iRes = _channel4.showInfo(getDeviceIdRemote(), 0, 0, 0, 0, 0, szMessage);

        return resultToError(iRes, piError);}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::sendMessage(short deviceIDRemote, unsigned char ucDelayTime, 
                            unsigned char ucAction,  unsigned char ucAudio, 
                            char *szMessage, int iLength, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2: {
        unsigned char Buffer[245];
        int counter;

        Buffer[0] = ucDelayTime;
        Buffer[1] = ucAction;
        Buffer[2] = ucAudio;

        for(counter=0; (counter < iLength) && (counter < (int)(sizeof(Buffer) - 3)); counter++)
        {			
            Buffer[3 + counter] = szMessage[counter];
        }

        if(!_channel1_2.WriteMem(DISPLAY_MEMORY, counter + 3, &Buffer[0], piError))
        {			
            return 0;
        }
        return 1;}

    case 4: {
        int iRes = IV_ERROR_SUCCESS;

        // Befehlsausfuehrung starten.
        iRes = _channel4.showMessage(getDeviceIdRemote(), ucDelayTime, ucAction, ucAudio, szMessage);

        return resultToError(iRes, piError);}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Datensatzbeschreibung oder Listenbeschreibung aus dem Geraet lesen.
int ChannelBridge::getTableDescription(short deviceIdRemote, unsigned char ucList, int *piError)
{
    int res = 0;
    // Busnummer uebernehmen
    setDeviceIdRemote( deviceIdRemote );

    // Für den späteren Zugriff auf die Daten des Puffers die ProtocolVersion halten.
    unsigned char protocolVersion = getProtocolVersion( deviceIdRemote );

    switch ( protocolVersion )
    {
    case 1:  // AEIII+
    case 2:  // PZE - Master IV        
        res = _channel1_2.GetTables( ucList, piError );
        break;

    case 4:
        // Befehlsausfuehrung starten.
        res = resultToError( _channel4.getTablesWithTimeboySlot( deviceIdRemote, (ucList) ? (char*)"LIST" : (char*)"RECORD" ), piError );
        break;

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    if ( res )
    {
        if ( ucList )
            _protocolVersionListDescription = protocolVersion;
        else
            _protocolVersionRecordDescription = protocolVersion;
    }

    return res;
}

// Listenbeschreibung aus dem Geraet lesen.
int ChannelBridge::getListName(short deviceIDRemote, unsigned char ucList, char *name, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:
        return _channel1_2.GetListName(ucList, name, piError);
        break;

    case 4:
        *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
        return 0;

    default:
        *piError = IV_ERROR_UNREACHABLE;
        return 0;
    }

    return 0;
}

int ChannelBridge::writeStudioFile(short deviceIdRemote, char *szFileName, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIdRemote);

    switch (getProtocolVersion(getDeviceIdRemote())) {
    case 1: 
    case 2:
        return _channel1_2.WriteSetupData(szFileName, piError);

    case 4: { // PZEMasterIV mit Protokollversion 4
        int iRes = IV_ERROR_SUCCESS;

        // Befehlsausfuehrung starten.
        iRes = _channel4.writeSetupWithTimeboySlot( getDeviceIdRemote(), szFileName );

        return resultToError(iRes, piError);}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::readStudioFile(short deviceIDRemote, char *szFileName, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    int res = 0;
    switch (getProtocolVersion(getDeviceIdRemote())) {
    case 1: 
    case 2:
        return _channel1_2.ReadSetupData(szFileName, piError);

    case 4: // PZEMasterIV mit Protokollversion 4
        return resultToError(_channel4.readSetupWithTimeboySlot(getDeviceIdRemote(), szFileName), piError);

    default:
        *piError = IV_ERROR_UNREACHABLE;
        res = 0;
    }

    return res;
}

int ChannelBridge::modifyStudioFile(char *sourceFile, char *destFile, char *changeData, int flags, int *piError)
{
    (void)sourceFile;
    (void)destFile;
    (void)changeData;
    (void)flags;
    (void)piError;

#if !defined(DATAFOX_STUDIO_IV)
    CDFFileStudioXML file;
    int fileError = 0;
    struct stat st;

    if (stat(sourceFile, &st) != 0)
    {		
        // Datei konnte nicht ermittelt werden
        *piError = IV_ERROR_FILE_OPEN;
        return 0;
    }

    // Reset der StudioFile-Klasse
    file.SetDefault();
    file.SetLibraryVersion( libraryVersion );

    // Laden der Dateidaten. Keine Beachtung der Studioversion.
    if (!file.Load(sourceFile, &fileError, FILESTUDIOHEADER_IGNORE_MIN_STUDIO_VERSION | FILESTUDIOHEADER_IGNORE_SHOULD_STUDIO_VERSION)) {
        switch (fileError) {
        case -3:
            *piError = IV_ERROR_FILE_READ;
            break;
        case -7: // Mindestversion DLL wird nicht eingehalten.
            *piError = IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION;
            break;
        case -8: // Sollteversion DLL
            *piError = IV_ERROR_SETUPFILE_SHOULD_LIBRARY_VERSION;
            break;
        case -9: // Format der Setupdatei
            *piError = IV_ERROR_FILE_FORMAT;
            break;
        case -4: // CRC in Dateidaten
        case -10: case -11: case -12: case -13: case -14: case -15: case -16: case -17: case -18:
            *piError = IV_ERROR_FILE_CRC;
            break;
        default:
            *piError = IV_ERROR_FILE_OPEN;
        }
        return 0;
    }

    // StudioDatei modifizieren
    if (file.Modify(changeData, piError))
    {
        // StudioDatei speichern
        if (file.Save(destFile, &fileError))
        {
            // Alles erfolgreich durchgefuehrt
            return 1;
        }
        else
        {
            // Datei konnte nicht erstellt werden.
            *piError = IV_ERROR_FILE_OPEN;
        }
    }
#endif
    return 0;
}

int ChannelBridge::checkDevice(short deviceIDRemote, int *piError, int iPollRetry)
{
    int iRes = IV_ERROR_SUCCESS;

    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:
        return _channel1_2.CheckDevice( iPollRetry, piError );

    case 4: { // PZEMasterIV mit Protokollversion 4
        char szVersion[32];

        // Befehlsausfuehrung starten.
        iRes = _channel4.getVersion(getDeviceIdRemote(), szVersion);

        return resultToError(iRes, piError);}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    // Letzter Fehler ist aktueller
    return 0;
}

int ChannelBridge::restart(short deviceIDRemote, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 2:
        return _channel1_2.Restart( piError );

    case 1:
    case 4:
        *piError = IV_ERROR_VERSION_FIRMWARE;
        break;

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    // Letzter Fehler ist aktueller
    return 0;
}

int ChannelBridge::reset(short deviceIDRemote, int mode, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    int iRes = IV_ERROR_SUCCESS;
    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:
        *piError = IV_ERROR_VERSION_FIRMWARE;
        break;

    case 4:
        // Befehlsausfuehrung starten.
        iRes =  _channel4.systemReset(deviceIDRemote, mode);

        iRes = resultToError(iRes, piError);
        break;

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    // Letzter Fehler ist aktueller
    return iRes;
}

int ChannelBridge::getProtocolVersion2(short deviceIDRemote, unsigned char detailed, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    unsigned char Buf[4];

    // Protokollverwendung pruefen
    unsigned char pv = getProtocolVersion(getDeviceIdRemote());
    switch(pv)
    {
    case 1:
    case 2:
        // Detailierte Version auslesen.
        if (detailed)
        {
            // Version fuer Setup lesen
            if (!_channel1_2.ReadMem(MB_GET_VERSION, 2, Buf, piError))
            {
                return 0;
            }
        }
        else 
        {
            if (pv == 1)
            {
                // Es wird angenommen ein AEIII+ mit neuester Com-Version
                Buf[0] = 0x04;
                Buf[1] = 0x02;
            }
            else
            {
                // Es wird angenommen ein PZE Master IV mit neuester Com-Version
                Buf[0] = 0x05;
                Buf[1] = 0x01;
            }
        }
        break;

    case 3: 
        // Ab dem Protokolltyp 3 wird nur noch ueber die Firmwareversion entschieden.
        Buf[0] = 0x06;
        Buf[1] = 0x00;
        break;

    case 4:
        // Ab dem Protokolltyp 4 wird nur noch ueber die Firmwareversion entschieden.
        Buf[0] = 0x07;
        Buf[1] = 0x00;
        break;

    default:
        *piError = IV_ERROR_UNREACHABLE;
        return 0;
    }


    // Version fuer Kommunikation

    // Ab Firmware 04.01.00 Protokollanpassung (Version 4) und Listen in Tabellenart abhandeln.
    if ((Buf[0] == 7 && Buf[1] == 0) ||
        // Ab Firmware 04.01.00 Protokollanpassung (Version 3)
        (Buf[0] == 6 && Buf[1] == 0) ||
        // Ab Firmware 04.00.23 Listengroesse auf 256kByte erhoeht.
        (Buf[0] == 5 && Buf[1] == 2) ||
        // Ab Firmware 04.00.21 Listen im Hintergrund laden
        (Buf[0] == 5 && Buf[1] == 1) ||
        // Ab Firmware 04.00.19 des PZEMasterIV (Version 2)
        (Buf[0] == 5 && Buf[1] == 0) ||
        // Ab Firmware 03.01.21 wegen auslesen der Firmwareversion
        (Buf[0] == 4 && Buf[1] == 2) ||
        // Ab Firmware 03.00.12 mit Zutrittskontrolle
        (Buf[0] == 4 && Buf[1] == 1) ||
        // Ab Firmware 3.0 mit 16kByte Flash fuer Setupdaten und 16kByte Extended Flash
        (Buf[0] == 4 && Buf[1] == 0) )
    { 
        // Version als Rueckgabe
        return (Buf[0] << 8) | Buf[1];
    }

    *piError = IV_ERROR_VERSION_FIRMWARE;
    return 0;
}

int ChannelBridge::getSerialNum(short deviceIDRemote, int *piSerialNum, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1: 
    case 2:
        return _channel1_2.GetSerialNum( piSerialNum, piError );

    case 4: {
        int iRes;
        unsigned int serial;

        // Befehlsausfuehrung starten.
        iRes = _channel4.getSerial(getDeviceIdRemote(), serial);

        iRes = resultToError(iRes, piError);
        if (iRes)
        {
            // OK
            *piSerialNum = serial;
        }

        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// ListenBuffer in AE laden
int ChannelBridge::writeListData(short deviceIDRemote, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1: 
    case 2: 
        return _channel1_2.WriteListData( &_channel4, piError );

    case 4: // PZEMasterIV mit Protokollversion 4

        // Befehlsausfuehrung starten.
        return resultToError(_channel4.writeListWithTimeboySlot(getDeviceIdRemote()), piError);

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// ListenBuffer in AE laden
int ChannelBridge::writeEntranceListData(short deviceIDRemote, int iListNum, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    // Protokollklasse
    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 4:
        *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
        return 0;

    case 1:
    case 2:
        return _channel1_2.WriteEntranceListData( iListNum, piError );

    default:
        *piError = IV_ERROR_UNREACHABLE;
        return 0;
    }

    return 0;
}

// ListenBuffer in AE laden
int ChannelBridge::writeEntrance2ListData(short deviceIDRemote, int iListNum, int *piError)
{
    int iRes = IV_ERROR_SUCCESS;

    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:
        *piError = IV_ERROR_VERSION_FIRMWARE;
        return 0;

    case 4: 
        // Befehlsausfuehrung starten.
        // Listendaten zum AEIV uebertragen
        return resultToError(_channel4.writeEntrance2ListWithTimeboySlot(deviceIDRemote, iListNum), piError);

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// ListenBuffer in AE laden
int ChannelBridge::writeTimeboyListData(short deviceIDRemote, unsigned char id, unsigned char num, int reserve, int *piError)
{
    int iRes = IV_ERROR_SUCCESS;

    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:
        *piError = IV_ERROR_VERSION_FIRMWARE;
        return 0;

    case 4:
        // Befehlsausfuehrung starten.
        // Listendaten zum AEIV uebertragen
        return resultToError(_channel4.writeTimeboyListWithTimeboySlot(deviceIDRemote, id, num, reserve), piError);

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Abfrage ob Datensaetze im AEIII+ vorliegen.
// Erst ab Datensatzweisem lesen.
int ChannelBridge::checkData(short deviceIDRemote, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:  // AEIII+
    case 2:  // PZEMasterIV
        return _channel1_2.CheckData(piError);

    case 4: { // PZEMasterIV mit Protokollversion 4
        int iRes = IV_ERROR_SUCCESS;
        unsigned char buf[0x255];
        int count = 0;
        unsigned char status = 0;

        // Befehlsausfuehrung starten.
        iRes = _channel4.getRecord(getDeviceIdRemote(), buf, count, status);

        iRes = resultToError(iRes, piError);
        if (iRes)
        {
            // OK
            iRes = status;
        }
        else
        {
            iRes = -1;
        }
        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return -1;
}

// Schliesst Relais mit Nummer fuer Zeit.
int ChannelBridge::closeRelay(short deviceIDRemote, int iNum, int iTime, int *piError)
{
    int iRes = IV_ERROR_SUCCESS;
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:
        return _channel1_2.CloseRelay((unsigned char)getDeviceIdRemote(), iNum, iTime, piError);

    case 4: { // PZEMasterIV mit Protokollversion 4
        if (iNum > 4)
        {
            *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
            iRes = 0;
        }
        else
        {
            // Befehlsausfuehrung starten.
            iRes = resultToError(_channel4.setRelay(getDeviceIdRemote(), (char*)"INTERN", 0, iNum, 1, iTime), piError);
        }
        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Ermittelt aktuellen Zustand von Relais mit Nummer fuer Zeit.
int ChannelBridge::getRelayState(short deviceIDRemote, int iNum, int *piState, int *piTime, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:
        return _channel1_2.GetRelayState((unsigned char)getDeviceIdRemote(), iNum, piState, piTime, piError);

    case 4: { // PZEMasterIV mit Protokollversion 4
        int iRes = IV_ERROR_SUCCESS;
        if (iNum > 4)
        {
            *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
            iRes = 0;
        }
        else
        {		
            unsigned char state;
            unsigned short time;

            // Befehlsausfuehrung starten.
            iRes = resultToError(_channel4.getRelay(getDeviceIdRemote(), (char*)"INTERN", 0, iNum, state, time), piError);

            if (iRes)
            {
                // OK
                *piState = state;
                *piTime = time;
            }
        }
        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Oeffnet Relais mit Nummer.
int ChannelBridge::openRelay(short deviceIDRemote, int iNum, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    int iRes = IV_ERROR_SUCCESS;

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:
        return _channel1_2.OpenRelay((unsigned char)getDeviceIdRemote(), iNum, piError);

    case 4: { // PZEMasterIV mit Protokollversion 4
        if (iNum > 4)
        {
            *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
            iRes = 0;
        }
        else
        {		
            unsigned short time = 0;

            // Befehlsausfuehrung starten.
            iRes = resultToError(_channel4.setRelay(getDeviceIdRemote(), (char*)"INTERN", 0, iNum, 0, time), piError);
        }
        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Daten einer Variablen setzen.
int ChannelBridge::setGlobVar(short deviceIDRemote, unsigned char *pucVar, int iType, char *val, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    int iRes = IV_ERROR_SUCCESS;

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:
        return _channel1_2.SetGlobVar((unsigned char)getDeviceIdRemote(), pucVar, iType, val, piError);

    case 4: { // PZEMasterIV mit Protokollversion 4

        // Befehlsausfuehrung starten.
        switch (iType)
        {
        case 0: {
            char name[64];
            memset(name, 0, sizeof(name));
            sprintf(name, "SETUP.%i", *(unsigned char*)pucVar);
            iRes = _channel4.setVariable(getDeviceIdRemote(), name, val);
            break;}

        case 1: {
            char name[64];
            memset(name, 0, sizeof(name));
            strcpy(name, "SETUP.");
            strcat(name, (char*)pucVar);
            iRes = _channel4.setVariable(getDeviceIdRemote(), name, val);
            break;}

        case 2:
            iRes = _channel4.setSystemVariable(getDeviceIdRemote(), (char*)pucVar, val);
            break;
        }

        if (iRes > 0)
        {
            switch(iRes)
            {
            case 1:
            case 2:
                *piError = IV_ERROR_VARIABLE_UNDEFINED;
                iRes = 0;
                break;
            case 3:
                *piError = IV_ERROR_VARIABLE_VALUE;
                iRes = 0;
                break;
            default:
                iRes = resultToError(iRes, piError);
            }
        }
        else
        {
            iRes = resultToError(iRes, piError);
        }
        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Daten einer Globalen Variablen lesen.
int ChannelBridge::getGlobVar(short deviceIDRemote, unsigned char *pucVar, int iType, char *val, int iValueLength, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);
    char lValue[IV_PACKET_LENGTH_DATA];
    memset(lValue, 0, sizeof(lValue));

    // muss mindestens ein Zeichen anfordern.
    if ((iValueLength - 1) <= 0)
    {
        *piError = IV_ERROR_FUNCTION_PARAMETER;
        return 0;
    } 

    int iRes = IV_ERROR_SUCCESS;
    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:
        return _channel1_2.GetGlobVar((unsigned char)getDeviceIdRemote(), pucVar, iType, val, iValueLength, piError);

    case 4: { // PZEMasterIV mit Protokollversion 4

        // Befehlsausfuehrung starten.
        switch (iType) 
        {
        case 0: {
            char name[64];
            memset(name, 0, sizeof(name));
            sprintf(name, "SETUP.%i", *(unsigned char*)pucVar);
            iRes = _channel4.getVariable(deviceIDRemote, name, lValue);
            break;}

        case 1: {
            char name[64];
            memset(name, 0, sizeof(name));
            strcpy(name, "SETUP.");
            strcat(name, (char*)pucVar);
            iRes = _channel4.getVariable(deviceIDRemote, name, lValue);
            break;}

        case 2:
            iRes = _channel4.getSystemVariable(deviceIDRemote, (char*)pucVar, lValue);
            break;
        }

        if (iRes > 0)
        {
            switch ( iRes )
            {
            case 1:
            case 2:
                *piError = IV_ERROR_VARIABLE_UNDEFINED;
                iRes = 0;
                break;
            case 3:
                *piError = IV_ERROR_VARIABLE_VALUE;
                iRes = 0;
                break;
            default:
                iRes = resultToError(iRes, piError);
            }
        }
        else
        {
            iRes = resultToError(iRes, piError);
        }
        if (iRes) {
            strncpy(val, lValue, iValueLength - 1);
        }

        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::getVersionFirmware(short deviceIDRemote, char *buf, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    unsigned char lBuf[32];

    unsigned char pv = getProtocolVersion(getDeviceIdRemote());
    switch (pv)
    {
    case 1: 
    case 2: 
        return _channel1_2.GetVersionFirmware(buf, piError);

    case 4: {
        int iRes = IV_ERROR_SUCCESS;

        // Befehlsausfuehrung starten.
        iRes = resultToError(_channel4.getVersion(getDeviceIdRemote(), (char*)lBuf), piError);

        if (iRes)
        {
            // OK
            strcpy(buf, (char*)lBuf);
        }
        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::getVersionFirmwareFromFile(char *fileName, char *versionFirmware, int *error)
{
    return resultToError(_channel4.getVersionFirmwareFromFile(fileName, versionFirmware), error);
}

// Prueft ob Displaydaten in Ordnung.
int ChannelBridge::checkFileDisplay(FileDisplay *pFileDisplay, int *piError)
{
    return _channel1_2.CheckFileDisplay(pFileDisplay, piError);
}

int ChannelBridge::writeTextData(short deviceIDRemote, char *fileName)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.setTextDataWithTimeboySlot(deviceIDRemote, fileName);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeTouchKeyData(short deviceIDRemote, char *fileName)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.setTouchKeyDataWithTimeboySlot(deviceIDRemote, fileName);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeDisplayColors(short deviceIDRemote, char *fileName)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.setDisplayColorsWithTimeboySlot(deviceIDRemote, fileName);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeWlanConfigData(short deviceIDRemote, char *fileName)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.setWlanConfigDataWithTimeboySlot(deviceIDRemote, fileName);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::readWlanConfigData(short deviceIDRemote, char *fileName)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getWlanConfigDataWithTimeboySlot(deviceIDRemote, fileName);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeHardwareConfigData(short deviceIDRemote, char *fileName)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.setHardwareConfigDataWithTimeboySlot(deviceIDRemote, fileName);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::readHardwareConfigData(short deviceIDRemote, char *fileName)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getHardwareConfigDataWithTimeboySlot(deviceIDRemote, fileName);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeFileData(short deviceIDRemote, char *fileName)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.writeFileDataWithTimeboySlot(deviceIDRemote, fileName);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::readFileData(short deviceIDRemote, char *fileName)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.readFileDataWithTimeboySlot(deviceIDRemote, fileName);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeFileDevice(short deviceIDRemote, char *szFileName, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    int iRes = 0;
    switch (getProtocolVersion(deviceIDRemote)) 
    {
    case 2: // Protokollversion 2 gibt es nur beim PZE-MasterIV
        iRes = _channel1_2.WriteFileDevice(szFileName, piError);
        break;

    case 4: // MasterIV mit Protokollversion 4

        // Befehlsausfuehrung starten.
        iRes = resultToError(_channel4.writeDeviceWithTimeboySlot(deviceIDRemote, szFileName), piError);
        break;
    }

    if (iRes)
    {
        // Protokollversion ist aufgrund des Updates auf 0 zu setzen.
        // Naechste Kommunikationsanfage ermittelt die Protokollversion neu.
        _protocolVersion[(getDeviceIdRemote() == 0xFE) ? 0 : getDeviceIdRemote()] = 0;
    }

    return 0;
}

int ChannelBridge::writeModule(short deviceIDRemote, int type, char *szFileName, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return resultToError(_channel4.writeModuleWithTimeboySlot(deviceIDRemote, type, szFileName), piError);
    }

    *piError = IV_ERROR_UNREACHABLE;    // Zeichensatz- und Grafikdaten
    return 0;
}

//#########################################################
//######## Schnittstelle zu Channel1/Channel2 #############
//#########################################################

int ChannelBridge::getModus(short deviceIDRemote, int *error)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 4: return -IV_ERROR_VERSION_FIRMWARE;
    case 2: return _channel1_2.GetModus(error);
    }

    return -IV_ERROR_UNREACHABLE;
}


//#########################################################
//############# Schnittstelle zu Channel4 #################
//#########################################################

void ChannelBridge::setCallBackFunctionWithParams(int (*pFunction)(unsigned int, unsigned int))
{
    _channel4.setCallBackFunctionWithParams(pFunction);
}

unsigned int ChannelBridge::crc32(unsigned char *data, unsigned int size, unsigned int crc, unsigned char closure)
{
    return _channel4.crc32(data, size, crc, closure);
}

// MODULE_HARDWARE

// Einfache Anfrage ob das Geraet antwortet.
int ChannelBridge::ping(short deviceIDRemote)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.ping(deviceIDRemote, 0, 0);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeDevice(short deviceIDRemote, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getTypeDevice(deviceIDRemote, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeBoard(short deviceIDRemote, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getTypeBoard(deviceIDRemote, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentRS485_0(short deviceIDRemote, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getComponentRS485_0(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeTCPIP(short deviceIDRemote, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getTypeTCPIP(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeModemMC(short deviceIDRemote, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getTypeModemMC(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentRS485_1(short deviceIDRemote, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getComponentRS485_1(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeDisplay(short deviceIDRemote, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getTypeDisplay(deviceIDRemote, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeKeypad(short deviceIDRemote, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getTypeKeypad(deviceIDRemote, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeTransponder(short deviceIDRemote, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getTypeTransponder(deviceIDRemote, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeMMC(short deviceIDRemote, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getTypeMMC(deviceIDRemote, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeRadio(short deviceIDRemote, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getTypeRadio(deviceIDRemote, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeGPS(short deviceIDRemote, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getTypeGPS(deviceIDRemote, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentUSB(short deviceIDRemote, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getComponentUSB(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeDCF77(short deviceIDRemote, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getTypeDCF77(deviceIDRemote, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentInAnalog(short deviceIDRemote, unsigned int &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getComponentInAnalog(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentInDigital(short deviceIDRemote, unsigned int &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getComponentInDigital(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentOutDigital(short deviceIDRemote, unsigned int &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getComponentOutDigital(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeFingerprint(short deviceIDRemote, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getTypeFingerprint(deviceIDRemote, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeExternalReader(short deviceIDRemote, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getTypeExternalReader(deviceIDRemote, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeTouch(short deviceIDRemote, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getTypeTouch(deviceIDRemote, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeSensor(short deviceIDRemote, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getTypeSensor(deviceIDRemote, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentBattery(short deviceIDRemote, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getComponentBattery(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentCharge(short deviceIDRemote, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getComponentCharge(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionIOModule(short deviceIDRemote, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getOptionIOModule(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionTBAdaption(short deviceIDRemote, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getOptionTBAdaption(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionComServer(short deviceIDRemote, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getOptionComServer(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionTPWrite(short deviceIDRemote, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getOptionTPWrite(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionHTTP(short deviceIDRemote, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getOptionHTTP(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionLanguage(short deviceIDRemote, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getOptionLanguage(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionBiokeyUpdate(short deviceIDRemote, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getOptionBiokeyUpdate(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setSetupOptions(short deviceID, unsigned short mask, unsigned short options)
{
    switch (getProtocolVersion(deviceID))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.setSetupOptions(deviceID, mask, options);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getSetupOptions(short deviceIDRemote, unsigned short &mask, unsigned short &options)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getSetupOptions(deviceIDRemote, mask, options);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getPayableOptions(short deviceIDRemote, unsigned char &mask, unsigned char &options)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getPayableOptions(deviceIDRemote, mask, options);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getHWConfig(short deviceIDRemote, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getHWConfig(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getHardwareInfo(short deviceIDRemote, StructHardwareInfo &hwInfo)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.Channel4ModuleTable::getHardwareInfo(deviceIDRemote, hwInfo);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getHardwareInfo( short deviceIDRemote, HardwareInfo &hardwareInfo )
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getHardwareInfo( deviceIDRemote, hardwareInfo );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getInfos(StructHardwareInfo hwInfo, string &info)
{
    return _channel4.getInfos(hwInfo, info);
}

// MODULE_SYSTEM

// Systemreset; Speicherbereiche loeschen
//int systemReset(short deviceIDRemote, unsigned short modus);

// Firmwareversion lesen
int ChannelBridge::getVersionFirmware(short deviceIDRemote, char *version)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getVersion(deviceIDRemote, version);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getVersionBootloader(short deviceIDRemote, char *version)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1: return -IV_ERROR_VERSION_FIRMWARE;
    case 2: 
    {
        int res, error;
        res = _channel1_2.GetVersionBootloader( version, &error );
        if (!res)
            return -error;
    }
    case 4: return _channel4.getVersionBootloader(deviceIDRemote, version);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getSerialNumber(short deviceIDRemote, unsigned int &serial)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getSerial(deviceIDRemote, serial);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::openComServerMode(short deviceIDRemote, char *comm, unsigned int baudRate, unsigned char byteSize, unsigned char parity, unsigned char stopBits, unsigned short flags, unsigned short timeoutToClose)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2:return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.openComServerMode(deviceIDRemote, comm, baudRate, byteSize, parity, stopBits, flags, timeoutToClose);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::closeComServerMode(short deviceIDRemote)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    return _channel4.closeComServerMode(deviceIDRemote);
}

// Datensegmentuebertragung beginnen/enden
int ChannelBridge::beginSegment(short deviceIDRemote, unsigned short art, unsigned int &address, unsigned int size, unsigned int crc)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.beginSetSegment(deviceIDRemote, art, address, size, crc);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::endSegment(short deviceIDRemote, unsigned short art, unsigned int crc)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.endSetSegment(deviceIDRemote, art, crc);
    }

    return -IV_ERROR_UNREACHABLE;
}


// Uebertragen von Speicherdaten in Datei.
int ChannelBridge::writeMemToFile(short deviceIdRemote, int type, int address, int size, char *fileName)
{
    // Remoteadresse einstellen
    setDeviceIdRemote(deviceIdRemote);

    // Aufgrund Protokollversion entscheiden
    // Protokollverwendung pruefen
    switch (getProtocolVersion(deviceIdRemote))
    {
    case 1:
    case 2:
        return _channel1_2.WriteMemToFile(type, address, size, fileName);

    case 4:
        return _channel4.writeMemToFile(deviceIdRemote, type, address, size, fileName);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Uebertragen von Speicherdaten aus einer Datei.
int ChannelBridge::writeFileToMem(short deviceIDRemote, char *fileName)
{
    // Remoteadresse einstellen
    setDeviceIdRemote(deviceIDRemote);

    // Aufgrund Protokollversion entscheiden
    // Protokollverwendung pruefen
    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: 
        return _channel1_2.WriteFileToMem(fileName);

    case 4:
        return _channel4.writeFileToMem(deviceIDRemote, fileName);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Flash setzen/lesen (die ersten 16kByte Bootloader koennen nur gelesen werden)
int ChannelBridge::setFlash(short deviceIDRemote, unsigned int address, unsigned char *buf, int count)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    int err;
    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1: return _channel1_2.WriteMem(address, count, buf, &err, 32);
    case 2: return _channel1_2.WriteMem(address, count, buf, &err);
    case 4: return _channel4.setFlash(deviceIDRemote, address, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getFlash(short deviceIDRemote, unsigned int address, unsigned char *buf, int count)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    int err;
    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1: return _channel1_2.ReadMem(address, count, buf, &err);
    case 2: return _channel1_2.ReadMem(address, count, buf, &err);
    case 4: return _channel4.getFlash(deviceIDRemote, address, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getFlashPage(short deviceIDRemote, unsigned short pageNum, unsigned short offs, unsigned char *buf, int &count)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1: 
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getFlashPage(deviceIDRemote, pageNum, offs, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

// EEPROM setzen/lesen 
int ChannelBridge::setEEPROM(short deviceIDRemote, unsigned int address, unsigned char *buf, int count)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.setEEPROM(deviceIDRemote, address, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getEEPROM(short deviceIDRemote, unsigned int address, unsigned char *buf, int count)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getEEPROM(deviceIDRemote, address, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

// RAM setzen/lesen 
int ChannelBridge::setRAM(short deviceIDRemote, unsigned int address, unsigned char *buf, int count)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.setRAM(deviceIDRemote, address, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getRAM(short deviceIDRemote, unsigned int address, unsigned char *buf, int count)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return _channel4.getRAM(deviceIDRemote, address, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

// MMC setzen/lesen 
int ChannelBridge::setMMC(short deviceIDRemote, unsigned int address, unsigned char *buf, int count)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1: 
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return _channel4.setMMC(deviceIDRemote, address, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getMMC(short deviceIDRemote, unsigned int address, unsigned char *buf, int count)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getMMC(deviceIDRemote, address, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Systembios aufrufen/verlassen
int ChannelBridge::enterBios(short deviceIDRemote, unsigned char protocolVersion)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (protocolVersion)
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.enterBios(deviceIDRemote);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::exitBios(short deviceIDRemote)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.exitBios(deviceIDRemote);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Interface setzen/lesen (RS232, TCPIP...)
int ChannelBridge::setInterface(short deviceIDRemote, unsigned char type)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.setInterface(deviceIDRemote, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getInterface(short deviceIDRemote, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getInterface(deviceIDRemote, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Baud-Rate setzen/lesen
int ChannelBridge::setBaud(short deviceIDRemote, unsigned int br)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.setBaud(deviceIDRemote, br);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getBaud(short deviceIDRemote, unsigned int &br)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getBaudRate(deviceIDRemote, br);
    }

    return -IV_ERROR_UNREACHABLE;
}

// DeviceID setzen/lesen
int ChannelBridge::setDeviceIdBios(short deviceIDRemote, unsigned short id)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.setDeviceIdBios(deviceIDRemote, id);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getDeviceIdBios(short deviceIDRemote, unsigned short &id)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getDeviceIdBios(deviceIDRemote, id);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Bootloader aktivieren
int ChannelBridge::enterBootloader(short deviceIDRemote, unsigned char protocolVersion)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (protocolVersion)
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.enterBootloader(deviceIDRemote);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Monitor Mode vom XPort
int ChannelBridge::enterXPortMonitor(short deviceIDRemote, char *data, int &count)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.enterXPortMonitor(deviceIDRemote, data, count);
    }

    //return -IV_ERROR_UNREACHABLE;
    return -IV_ERROR_ENTER_MONITOR;
}

// Monitor Mode vom XPort
int ChannelBridge::enterXPortMonitorExt(short deviceIDRemote, char *data, int &count)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.enterXPortMonitorExt(deviceIDRemote, data, count);
    }

    //return -IV_ERROR_UNREACHABLE;
    return -IV_ERROR_ENTER_MONITOR;
}

int ChannelBridge::commandXPort(short deviceIDRemote, char *buf, int &count)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.commandXPort(deviceIDRemote, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::exitXPortMonitor(short deviceIDRemote)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.exitXPortMonitor(deviceIDRemote);
    }

    //return -IV_ERROR_UNREACHABLE;
    return -IV_ERROR_EXIT_MONITOR;
}

// Softwareinformation der Anwendungsfirmware abfragen
int ChannelBridge::getSoftwareInfo(short deviceIDRemote, unsigned char *swi, int &infoLength)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.Channel4ModuleTable::getSoftwareInfo(deviceIDRemote, swi, infoLength);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Oeffnen einer Tabelle
int ChannelBridge::tableOpen(short deviceIDRemote, char *name, int options, int &handle, char *reserved)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.tableOpen(deviceIDRemote, name, options, handle, reserved);
    }

    return -IV_ERROR_UNREACHABLE;
}
// Schliessen der Tabelle
int ChannelBridge::tableClose(short deviceIDRemote, int handle)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.tableClose(deviceIDRemote, handle);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Setzen eines Filters auf die Tabelle
int ChannelBridge::tableSetFilter(short deviceIDRemote, int handle, char *filter)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.tableSetFilter(deviceIDRemote, handle, filter);
    }

    return -IV_ERROR_UNREACHABLE;
}
// Lesen des Filters auf die Tabelle
int ChannelBridge::tableGetFilter(short deviceIDRemote, int handle, char *filter, short &len)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.tableGetFilter(deviceIDRemote, handle, filter, len);
    }

    return -IV_ERROR_UNREACHABLE;
}
// Loeschen des Filters der Tabelle
int ChannelBridge::tableClearFilter(short deviceIDRemote, int handle)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.tableClearFilter(deviceIDRemote, handle);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Ueberspringen von n Datensaetzen (offset kann negativ sein), ab top/aktuell/bottom
int ChannelBridge::tableSkip(short deviceIDRemote, int handle, int offset, short origin)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.tableSkip(deviceIDRemote, handle, offset, origin);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Aendern eines Spaltenwerts, des aktuellen Datensatzes
int ChannelBridge::tableSetField(short deviceIDRemote, int handle, char *name, char *val)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.tableSetField(deviceIDRemote, handle, name, val);
    }

    return -IV_ERROR_UNREACHABLE;
}
// Abfragen eines Spaltenwerts, des aktuellen Datensatzes
int ChannelBridge::tableGetField(short deviceIDRemote, int handle, char *name, char *val, short &len)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.tableGetField(deviceIDRemote, handle, name, val, len);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Abragen eines Modulinfos
int ChannelBridge::getInfo(short deviceIDRemote, char *name, char *params, char *info, int &len)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.getInfo(deviceIDRemote, name, params, info, len);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Zutrittsbuchung durchfuehren
int ChannelBridge::entrance2Identification(short deviceIDRemote, char *TM, char *Id, char *Pin, char *status, int &statusLength)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.Entrance2Identification(deviceIDRemote, TM, Id, Pin, status, statusLength);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::entrance2OnlineAction(short deviceIDRemote, char *TM, unsigned int mask, unsigned char type, unsigned char duration)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.Entrance2OnlineAction(deviceIDRemote, TM, mask, type, duration);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setNotifyOnDevice(short deviceIDRemote, int type, int state)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.setNotify(deviceIDRemote, type, state);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Datensatz an Fingerprint-Tabelle anfuegen.
int ChannelBridge::fingerprintAppendRecord(short deviceIDRemote, unsigned char type, unsigned char *fingertemplate, int size)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.fingerprintTableAppendRecord(deviceIDRemote, type, fingertemplate, size);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Datensatz aus Fingerprint-Tabelle lesen.
int ChannelBridge::fingerprintGetRecord(short deviceIDRemote, unsigned char &type, int pid, int fid, unsigned char *fingertemplate, int &size)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.fingerprintTableGetRecord(deviceIDRemote, type, pid, fid, fingertemplate, size);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Datensatz aus Fingerprint-Tabelle loeschen.
int ChannelBridge::fingerprintDeleteRecord(short deviceIDRemote, int pid, int fid)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.fingerprintTableDeleteRecord(deviceIDRemote, pid, fid);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Liste mit Fingerprintmoduldaten erzeugen.
int ChannelBridge::fingerprintList(short deviceIDRemote, int flags, int &recordCount)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.fingerprintList(deviceIDRemote, flags, recordCount);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Backup der Fingerprints erstellen
int ChannelBridge::fingerprintBackup(short deviceIDRemote, char *backupFile, int flags)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.fingerprintBackup(deviceIDRemote, backupFile, flags);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Flags:
// Bit 1: gesetzt bedeutet: Enthaltene Fingertemplates beibehalten.
// Wiederherstellen der Fingerprints
int ChannelBridge::fingerprintRestore(short deviceIDRemote, char *backupFile, int flags)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.fingerprintRestore(deviceIDRemote, backupFile, flags);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Baud-Rate setzen/lesen
int ChannelBridge::setSystemFlags(short deviceIDRemote, unsigned int flags)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return _channel4.setSystemFlags(deviceIDRemote, flags);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getSystemFlags(short deviceIDRemote, unsigned int &flags)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return _channel4.getSystemFlags(deviceIDRemote, flags);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setOptionFirmware(short deviceIDRemote, int mask, int options)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return _channel4.setOptionFirmware(deviceIDRemote, mask, options);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionFirmware(short deviceIDRemote, int &mask, int &options)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return _channel4.getOptionFirmware(deviceIDRemote, mask, options);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Setzt den zu verwendenden Zeichensatztyp fuer sendMessage oder Sendinfotext
int ChannelBridge::setTypeFont(short deviceIDRemote, int iType, int *piError)
{
    int iRes = IV_ERROR_SUCCESS;
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:
        *piError = IV_ERROR_VERSION_FIRMWARE;
        break;

    case 4: // PZEMasterIV mit Protokollversion 4
        // Befehlsausfuehrung starten.
        return resultToError(_channel4.SetTypeFont(getDeviceIdRemote(), iType), piError);

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Passwort pruefen
int ChannelBridge::passwordCheck(short deviceIDRemote, char *passwort)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.passwordCheck(deviceIDRemote, passwort);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Abragen eines Modulinfos
int ChannelBridge::getPasswordKey(short deviceIDRemote, char *key)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.passwordKey(deviceIDRemote, key);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Tastendruck ueber Kommunikationsbefehl simulieren
int ChannelBridge::pressVirtualKey(short deviceIDRemote, unsigned char vk, unsigned char flags, int *piError)
{
    int iRes = IV_ERROR_SUCCESS;
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:
        *piError = IV_ERROR_VERSION_FIRMWARE;
        break;

    case 4: // PZEMasterIV mit Protokollversion 4
        // Befehlsausfuehrung starten.
        return resultToError(_channel4.pressVirtualKey(getDeviceIdRemote(), vk, flags), piError);

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::getFlashStatus(short deviceIDRemote, int *piStatus, int *piError)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    unsigned int serial = 0;

    int iRes = 0;
    switch (getProtocolVersion(getDeviceIdRemote()))
    {
    case 1:
    case 2:
        *piError = IV_ERROR_VERSION_FIRMWARE;
        break;

    case 4:

        iRes = resultToError(_channel4.getSerial(getDeviceIdRemote(), serial), piError);
        if (iRes)
        {
            // Flags in Statusvariable uebernehmen
            *piStatus = (_channel4._lastFlags & 0x70) >> 4;
        }
        break;
    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    // Letzter Fehler ist aktueller
    return iRes;
}

// Passwort fuer die verschluesselte Kommunikation
int ChannelBridge::setCommunicationPassword(unsigned char *password, int size, int sessionTimeout, int global)
{
    // Aufruf an CChannel4 weiterleiten
    return _channel4.setCommunicationPassword(password, size, sessionTimeout, global);
}

// Pruefen ob eine Globale Passwortvergabe moeglich ist
int ChannelBridge::isGlobalCommunicationPassword()
{
    // Aufruf an CChannel4 weiterleiten
    return _channel4.getCryptSession().IsGlobalCommunicationPassword();
}

int ChannelBridge::recordVolume(short deviceID, int &number, int &memoryUsage, int &memorySize)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceID);

    switch (getProtocolVersion(deviceID))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.recordVolume(deviceID, number, memoryUsage, memorySize);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Funktionen für Pufferzugriff
int ChannelBridge::GetTableCount(unsigned char ucList)
{
    // Anhand der zuvor angesprochenen Protocolversion bei getTableDescription den weiteren Zugriff steuern.
    unsigned char protocolVersion = (ucList) ? _protocolVersionListDescription : _protocolVersionRecordDescription;

    switch ( protocolVersion )
    {
    case 1:
    case 2:
        return _channel1_2.GetTableCount( ucList );

    case 4:
        return _channel4.GetTableCount( ucList );
    }

    return 0;
}

int ChannelBridge::GetTableRecord(unsigned char ucNum,unsigned char *szName, unsigned char *pucFieldCount, unsigned char *pucVariable, unsigned char ucList)
{
    // Anhand der zuvor angesprochenen Protocolversion bei getTableDescription den weiteren Zugriff steuern.
    unsigned char protocolVersion = (ucList) ? _protocolVersionListDescription : _protocolVersionRecordDescription;

    // Bei Listen kann es nötig sein auf die Zutrittslistendaten zu verzweigen.
    if ( ucList && ucNum >= 100 )
    {
        // Auf Zutrtittskontrolllisten verzweigen
        if (ucNum >= 100 && ucNum < 104) {
            // Version 1
            *pucVariable = 0;
            return _channel1_2.GetTableRecordEntrance(ucNum, szName, pucFieldCount);
        } else if (ucNum >= 200 && ucNum < 208) {
            // Version 2
            *pucVariable = 0;
            return _channel4.GetTableRecordEntrance2(ucNum, szName, pucFieldCount);
        }
    }

    switch ( protocolVersion )
    {
    case 1:
    case 2:
        return _channel1_2.GetTableRecord( ucNum, szName, pucFieldCount, pucVariable, ucList );

    case 4:
        return _channel4.GetTableRecord( ucNum, szName, pucFieldCount, pucVariable, ucList );
    }

    return 0;
}

int ChannelBridge::GetTableField(unsigned char ucTableNum, unsigned char ucFieldNum, unsigned char *szName, unsigned char *pucType, unsigned char *pucLength, unsigned char ucList)
{
    // Anhand der zuvor angesprochenen Protocolversion bei getTableDescription den weiteren Zugriff steuern.
    unsigned char protocolVersion = (ucList) ? _protocolVersionListDescription : _protocolVersionRecordDescription;

    // Bei Listen kann es nötig sein auf die Zutrittslistendaten zu verzweigen.
    if ( ucList && ucTableNum >= 100 )
    {
        // Auf Zutrittskontrolllisten verzweigen
        if (ucTableNum >= 100 && ucTableNum < 104) {
            // Version 1
            return _channel1_2.GetTableFieldEntrance(ucTableNum, ucFieldNum, szName, pucType, pucLength);
        } else if (ucTableNum >= 200 && ucTableNum < 208) {
            // Version 2
            return _channel4.GetTableFieldEntrance2(ucTableNum, ucFieldNum, szName, pucType, pucLength);
        }
    }

    switch ( protocolVersion )
    {
    case 1:
    case 2:
        return _channel1_2.GetTableField( ucTableNum, ucFieldNum, szName, pucType, pucLength, ucList );

    case 4:
        return _channel4.GetTableField( ucTableNum, ucFieldNum, szName, pucType, pucLength, ucList );
    }

    return 0;
}

void ChannelBridge::ClearListBuffer()
{
    // Importiert wird nur in Channel4, bei Load wird ggf. von Channel1_2 auf Channel4Buffer zugegriffen.
    _channel4.ClearListBuffer();
}

int ChannelBridge::MakeListe(int ListNum, int ListCnt, int ListSize, unsigned char *Buf)
{
    // Importiert wird nur in Channel4, bei Load wird ggf. von Channel1_2 auf Channel4Buffer zugegriffen.
    return _channel4.MakeListe( ListNum, ListCnt, ListSize, Buf );
}

void ChannelBridge::ClearEntranceListBuffer( int iListNum )
{
    _channel1_2.ClearEntranceListBuffer(iListNum);
}

int ChannelBridge::MakeEntranceList( int ListNum, int ListCnt, int ListSize, unsigned char *Buf )
{
    return _channel1_2.MakeEntranceList( ListNum, ListCnt, ListSize, Buf );
}

void ChannelBridge::ClearEntrance2ListBuffer( int iListNum )
{
    _channel4.ClearEntrance2ListBuffer( iListNum );
}

int ChannelBridge::MakeEntrance2List( int ListNum, int ListCnt, int ListSize, unsigned char *Buf )
{
    return _channel4.MakeEntrance2List( ListNum, ListCnt, ListSize, Buf );
}

void ChannelBridge::ClearTimeboyListBuffer( unsigned char id, unsigned char num )
{
    _channel4.ClearTimeboyListBuffer( id, num );
}

int ChannelBridge::MakeTimeboyList( unsigned char id, unsigned char num, unsigned short recordCount, unsigned char *buf, unsigned int size, int flags, int *error )
{
    return _channel4.MakeTimeboyList( id, num, recordCount, buf, size, flags, error );
}

bool ChannelBridge::WriteListToBinFile( unsigned char listNum, const char *filePath )
{
    return _channel4.WriteListToBinFile( listNum, filePath );
}

bool ChannelBridge::WriteListToTextFile( unsigned char listNum, const char *filePath )
{
    return _channel4.WriteListToTextFile( listNum, filePath );
}

bool ChannelBridge::WriteListEntrance2ToBinFile( unsigned char listNum, const char *filePath )
{
    return _channel4.WriteListEntrance2ToBinFile( listNum, filePath );
}

bool ChannelBridge::WriteListEntrance2ToTextFile( unsigned char listNum, const char *filePath )
{
    return _channel4.WriteListEntrance2ToTextFile( listNum, filePath );
}

bool ChannelBridge::WriteListTimeboyToBinFile( unsigned char listId, unsigned char listNum, const char *filePath )
{
    return _channel4.WriteListTimeboyToBinFile( listId, listNum, filePath );
}

string ChannelBridge::GetFileDeviceFileName()
{
    return _channel4.m_fileDevice.m_fileName;
}

string ChannelBridge::GetFileDeviceError()
{
    return _channel4.m_fileDevice.m_errorExtended;
}

string ChannelBridge::GetFileDeviceInfo()
{
    return _channel4._info;
}

FileDevice &ChannelBridge::GetFileDevice()
{
    return _channel4.m_fileDevice;
}

#if !defined( DATAFOX_STUDIO_IV ) && !defined( DF_API_PRODUCTION )
// Funktions-Stumps

// MODULE_HARDWARE

int ChannelBridge::setTypeDevice(unsigned char deviceIDRemote, unsigned char type)
{
    (void)deviceIDRemote;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeBoard(unsigned char deviceIDRemote, unsigned char type)
{
    (void)deviceIDRemote;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeTCPIP(unsigned char deviceIDRemote, unsigned char val)
{
    (void)deviceIDRemote;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeModemMC(unsigned char deviceIDRemote, unsigned char val)
{
    (void)deviceIDRemote;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeDisplay(unsigned char deviceIDRemote, unsigned char type)
{
    (void)deviceIDRemote;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeKeypad(unsigned char deviceIDRemote, unsigned char type)
{
    (void)deviceIDRemote;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeTransponder(unsigned char deviceIDRemote, unsigned char type)
{
    (void)deviceIDRemote;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeMMC(unsigned char deviceIDRemote, unsigned char type)
{
    (void)deviceIDRemote;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeRadio(unsigned char deviceIDRemote, unsigned char type)
{
    (void)deviceIDRemote;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeGPS(unsigned char deviceIDRemote, unsigned char type)
{
    (void)deviceIDRemote;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeDCF77(unsigned char deviceIDRemote, unsigned char type)
{
    (void)deviceIDRemote;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeFingerprint(unsigned char deviceIDRemote, unsigned char type)
{
    (void)deviceIDRemote;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeExternalReader(unsigned char deviceIDRemote, unsigned char val)  
{ 
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return _channel4.setTypeExternalReader(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setComponentRS485_0(unsigned char deviceIDRemote, unsigned char val)
{
    (void)deviceIDRemote;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setComponentRS485_1(unsigned char deviceIDRemote, unsigned char val)
{
    (void)deviceIDRemote;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setComponentUSB(unsigned char deviceIDRemote, unsigned char val)
{
    (void)deviceIDRemote;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setComponentInAnalog(unsigned char deviceIDRemote, unsigned int val)
{
    (void)deviceIDRemote;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setComponentInDigital(unsigned char deviceIDRemote, unsigned int val)
{
    (void)deviceIDRemote;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setComponentOutDigital(unsigned char deviceIDRemote, unsigned int val)
{
    (void)deviceIDRemote;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setOptionIOModule(unsigned char deviceIDRemote, unsigned char val)      
{ 
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return _channel4.setOptionIOModule(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setOptionTBAdaption(unsigned char deviceIDRemote, unsigned char val)    
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return _channel4.setOptionTBAdaption(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;

}

int ChannelBridge::setOptionComServer(unsigned char deviceIDRemote, unsigned char val)     
{ 
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return _channel4.setOptionComServer(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setOptionTPWrite(unsigned char deviceIDRemote, unsigned char val)       
{ 
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return _channel4.setOptionTPWrite(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setOptionHTTP(unsigned char deviceIDRemote, unsigned char val)          
{ 
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return _channel4.setOptionHTTP(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setOptionLanguage(unsigned char deviceIDRemote, unsigned char val)
{
    (void)deviceIDRemote;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setOptionBiokeyUpdate(unsigned char deviceIDRemote, unsigned char val)  
{ 
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return _channel4.setOptionBiokeyUpdate(deviceIDRemote, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setPayableOptions(unsigned char deviceID, unsigned char mask, unsigned char options)
{
    (void)deviceID;
    (void)mask;
    (void)options;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setHWConfig(unsigned char deviceIDRemote, unsigned char val)
{
    (void)deviceIDRemote;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

// MODULE_SYSTEM

int ChannelBridge::setSerialNumber(unsigned char deviceIDRemote, unsigned int serial)
{
    (void)deviceIDRemote;
    (void)serial;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

// Zutrittsbuchung durchfuehren
int ChannelBridge::accessControlIdentification(short deviceIDRemote, int masterId, char *TM, char *Id, char *Pin, char *status, int &statusLength)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.accessControlIdentification(deviceIDRemote, masterId, TM, Id, Pin, status, statusLength);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::accessControlOnlineAction(short deviceIDRemote, int masterId, char *TM, unsigned int mask, unsigned char type, unsigned char duration)
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.accessControlOnlineAction(deviceIDRemote, masterId, TM, mask, type, duration);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Zutrittskontroll-Befehl für Schließzylinder
int ChannelBridge::accessControlKnobCommand( short deviceIDRemote, unsigned char knobCommand, unsigned short masterId, unsigned char knobId, char *params, unsigned char *data, int *sizeData )
{
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceIDRemote);

    switch (getProtocolVersion(deviceIDRemote))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return _channel4.accessControlKnobCommand( deviceIDRemote,knobCommand, masterId, knobId, params, data, sizeData );
    }

    return -IV_ERROR_UNREACHABLE;
}


// Benutzerdefinierter Befehl
int ChannelBridge::customCommand( short deviceIDRemote, unsigned int idendification, unsigned int customCommand, char *data, int *sizeData )
{
	// Busnummer uebernehmen
	setDeviceIdRemote(deviceIDRemote);

	switch (getProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return _channel4.customCommand( deviceIDRemote, idendification, customCommand, data, sizeData );
	}

	return -IV_ERROR_UNREACHABLE;
}


int ChannelBridge::startTransparencyMode( unsigned char deviceAddress, unsigned short id, unsigned short flags, unsigned int sizeReadBuffer, unsigned int sizeWriteBuffer, int timeout, int *errorNumber )  
{ 
    // Busnummer uebernehmen
    setDeviceIdRemote(deviceAddress);
    int iRes = 0;

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: 
        {
            *errorNumber = IV_ERROR_VERSION_FIRMWARE;
            break;
        }
    case 4: 
        {
            iRes = _channel4.startTransparencyMode( deviceAddress, id, flags, sizeReadBuffer, sizeWriteBuffer, timeout );
            if ( iRes != IV_ERROR_SUCCESS )
            {
                // Spezielle Funktionsfehler auswerten, ansonsten nach Standard auflösen.
                switch ( (iRes >> 8) & 0xFF )
                {
                case 1:
                    *errorNumber = IV_ERROR_ALLOC_MEM;
                    iRes = 0;
                    break;

                case 2:
                    *errorNumber = IV_ERROR_TIMEBOY_IV;
                    iRes = 0;
                    break;

                default:
                    iRes = resultToError( iRes, errorNumber );
                }
            }
            else
            {
                iRes = 1;
            }
            break;
        }

    default:
        *errorNumber = IV_ERROR_UNREACHABLE;
        break;
    }

    return iRes;
}

int ChannelBridge::stopTransparencyMode( unsigned char deviceAddress, int *errorNumber )  
{ 
    int iRes = 0;

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: 
        {
            *errorNumber = IV_ERROR_VERSION_FIRMWARE;
            break;
        }
    case 4: 
        {
            iRes = _channel4.stopTransparencyMode( deviceAddress );
            if ( iRes != IV_ERROR_SUCCESS )
            {
                iRes = resultToError( iRes, errorNumber );
            }
            else
            {
                iRes = 1;
            }
            break;
        }

    default:
        *errorNumber = IV_ERROR_UNREACHABLE;
        break;
    }

    return iRes;
}

int ChannelBridge::isTransparencyMode()  
{ 
    return _channel4.isTransparencyMode();
}

int ChannelBridge::ping( unsigned char deviceAddress, int attempts, int *errorNumber )  
{ 
    // Einen Versuch mehr.
    attempts++;

    for ( int i = 1; i < attempts; ++i )
    {
        if ( _channel4.ping( deviceAddress, NULL, NULL ) == IV_ERROR_SUCCESS )
            return i;
    }

    *errorNumber = IV_ERROR_UNREACHABLE;
    return 0;
}
#endif
