/**
 * @file  filedisplay.cpp
 * @brief Implementierung der Klasse FileDisplay.
 */
//////////////////////////////////////////////////////////////////////

#include "filedisplay.h"
#include <memory.h>
#include <stdio.h>

FileDisplay::FileDisplay()
{
	// CRC32 Initialisierung
	InitCRC32Table();

	// Ausgangseinstellung
	resetError();
	reset();
}

FileDisplay::~FileDisplay()
{
}

void FileDisplay::reset()
{
	// Basis Reset
	FileIntelHex::reset();

	// Member initialisieren
	
	// Groessen setzen
	m_sizeData   = 0;
	m_sizeFont   = 0;
	m_sizeGrafic = 0;

	// Speicher fuer Zeichensatz- und Grafikdaten loeschen
	memset(m_memFont, 0x00, MAX_MEM_FONT);
	memset(m_memGrafic, 0x00, MAX_MEM_GRAFIC);

	// Hilfsstrukturen aufsetzen
	m_pFontNode   = (FONTNODE*)m_memFont;
	m_pFontNode->version = 1;
	m_pFontNode->cnt     = 0;
	m_pFontNode->temp[0] = 0;
	m_pFontNode->temp[1] = 0;
	m_pFontNode->ptrFontArray = sizeof(FONTNODE);

	m_pGraficNode = (GRAFICNODE*)m_memGrafic;
	m_pGraficNode->version = 1;
	m_pGraficNode->cnt     = 0;
	m_pGraficNode->temp[0] = 0;
	m_pGraficNode->temp[1] = 0;
	m_pGraficNode->ptrGraficArray = sizeof(GRAFICNODE);
}

void FileDisplay::Default(int iFontCount, int iGraficCount)
{
	// Variablen fuer Groessenangabe
	m_sizeData   = 0;
	
	DefaultFont(iFontCount);
	DefaultGrafic(iGraficCount);
}

void FileDisplay::DefaultFont( unsigned char count )
{
	int i, j;

	// Variablen fuer Groessenangabe
	m_sizeFont   = 0;

	// Speicher fuer Zeichensatz- und Grafikdaten loeschen
	memset(m_memFont, 0x00, MAX_MEM_FONT);

	// 1.
	// Zeichensaetze init
	// Grundstruktur aufsetzen
	m_pFontNode->version = 1;
	m_pFontNode->cnt     = count;
	m_pFontNode->temp[0] = 0;
	m_pFontNode->temp[1] = 0;
	m_pFontNode->ptrFontArray = sizeof(FONTNODE);

	// Zeigerarray und Zeichensatzstrukturzeiger initialisieren
	i = m_pFontNode->cnt;
	for (j=0; j < i; j++) {
		
		// Zeigereintrag auf Zeichensatzstruktur
		*(unsigned int*)&m_memFont[m_pFontNode->ptrFontArray + j * sizeof(unsigned int)] = 
			m_pFontNode->ptrFontArray + i * sizeof(unsigned int) + j * sizeof(FONT);

		m_pFont[j] = (FONT*)&m_memFont[m_pFontNode->ptrFontArray + i * sizeof(unsigned int) + j * sizeof(FONT)];
	}

}

void FileDisplay::DefaultGrafic( unsigned char count )
{
	int i, j;

	// Variablen fuer Groessenangabe
	m_sizeGrafic = 0;

	// Speicher fuer Grafikdaten loeschen
	memset(m_memGrafic, 0x00, MAX_MEM_GRAFIC);

	// 2.
	// Grafiken init
	// Grundstruktur aufsetzen
	m_pGraficNode->version = 1;
	m_pGraficNode->cnt     = count;
	m_pGraficNode->temp[0] = 0;
	m_pGraficNode->temp[1] = 0;
	m_pGraficNode->ptrGraficArray = sizeof(GRAFICNODE);

	// Zeigerarray und Grafikstrukturzeiger initialisieren
	i = m_pGraficNode->cnt;
	for (j=0; j < i; j++) {
		
		// Zeigereintrag auf Grafikstruktur
		*(unsigned int*)&m_memGrafic[m_pGraficNode->ptrGraficArray + j * sizeof(unsigned int)] = 
			m_pGraficNode->ptrGraficArray + i * sizeof(unsigned int) + j * sizeof(GRAFIC);

		m_pGrafic[j] = (GRAFIC*)&m_memGrafic[m_pGraficNode->ptrGraficArray + i * sizeof(unsigned int) + j * sizeof(GRAFIC)];
	}

}

void FileDisplay::FontCorrect( int iNum ) 
{
	int j = iNum;

	// Leerzeichen ist so breit wie Doppelpunkt
	m_memFont[m_pFont[j]->ptrCharacter + (m_memFont[m_pFont[j]->ptrTable + (unsigned char)' '] - 1) * m_pFont[j]->bytesPerChar] =
		m_memFont[m_pFont[j]->ptrCharacter + (m_memFont[m_pFont[j]->ptrTable + (unsigned char)':'] - 1) * m_pFont[j]->bytesPerChar];

	// 64 Pixel - Zeichensatz
	if (j == 2) {

		char q;
		unsigned char width = 0;
		
		// Suche breitestes Zeichen
		for (q='0'; q <= '9'; q++) {
			if(m_memFont[m_pFont[j]->ptrCharacter + (m_memFont[m_pFont[j]->ptrTable + (unsigned char)q] - 1) * m_pFont[j]->bytesPerChar] > width) {

				width = m_memFont[m_pFont[j]->ptrCharacter + (m_memFont[m_pFont[j]->ptrTable + (unsigned char)q] - 1) * m_pFont[j]->bytesPerChar];
			}
		}
		// Alle Zeichen so breit wie breitestes
		for (q='0'; q <= '9'; q++) {
			
			m_memFont[m_pFont[j]->ptrCharacter + (m_memFont[m_pFont[j]->ptrTable + (unsigned char)q] - 1) * m_pFont[j]->bytesPerChar] = width;
		}
	}

	// Neuberechnung des 32bit CRC ueber Tabelle und Zeichensatzdaten. (muessen immer direkt hintereinander im Speicher liegen)
	m_pFont[j]->crc  = GetCRC32(&m_memFont[m_pFont[j]->ptrTable], 0x100 + m_pFont[j]->size);
}

int FileDisplay::Load( const std::string &filePath )
{
	// Fehlerobjekt zuruecksetzen
	resetError();

	// Umgebung zuruecksetzen
	reset();

	// Laden der Zeichensatzdaten aus Hexfile
	if (!LoadFont( filePath )) {

		return 0;
	}

	// Laden der Grafikdaten aus Hexfile
	if (!LoadGrafic( filePath )) {

		return 0;
	}

	// Gesamtdatenbyte
	m_sizeData = m_sizeFont + m_sizeGrafic;

	return 1;
}

int FileDisplay::LoadFont( const std::string &filePath ) 
{
	// Lesen der Hexdatei
	FILE *fp;
	char data[0x400];

	if ((fp = fopen(filePath.data(), "r")) == (FILE*)0) {
		// Oeffnen der Datei gescheitert
		setErrorState( FileOpen );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
		return 0;
	}

	// Daten eintragen
	unsigned char type;
	unsigned short address;

	while(!feof(fp) && (errorState() == None)) {
	
		if (fgets(data, 0x400, fp) != 0) {
			
			address = GetAddress(data);
			type    = GetType(data);

			switch (type) {

			case TypeDisplayFont:
				m_sizeFont += GetData(data, &m_memFont[address]);
				break;

			case TypeDisplayGrafic:
				break;

			case TypeEOF:
				break;
			}
		}
	}
	fclose(fp);	

	// Member durch gelesene Daten initialisieren und pruefen
	if (m_sizeFont) {
		
		int i, j;

		// 1.
		// Zeichensaetze pruefen

		// Zeigerarray und Zeichensatzstrukturzeiger initialisieren
		i = m_pFontNode->cnt;
		if (i > MAX_FONT) {

			setErrorState( Format );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
			return 0;
		}

		for (j=0; j < i; j++) {
			
			m_pFont[j] = (FONT*)&m_memFont[m_pFontNode->ptrFontArray + i * sizeof(unsigned int) + j * sizeof(FONT)];
		}

		// CRC32 pruefen
		for (j=0; j < i; j++) {
			
			if (m_pFont[j]->crc != GetCRC32(&m_memFont[m_pFont[j]->ptrTable], 0x100 + m_pFont[j]->size)) {
				
				setErrorState( CrcFail );
				setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
				return 0;
			}
		}

	} else {

		setErrorState( NoData );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
		return 0;
	}

	return 1;
}

int FileDisplay::LoadGrafic( const std::string &filePath ) 
{
	// Lesen der Hexdatei
	FILE *fp;
	char data[0x400];

	if ((fp = fopen(filePath.data(), "r")) == (FILE*)0) {
		// Oeffnen der Datei gescheitert
		setErrorState( FileOpen );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
		return 0;
	}

	// Daten eintragen
	unsigned char type;
	unsigned short address;

	while(!feof(fp) && (errorState() == None)) {
	
		if (fgets(data, 0x400, fp) != 0) {
			
			address = GetAddress(data);
			type    = GetType(data);

			switch (type) {

			case TypeDisplayFont:
				break;

			case TypeDisplayGrafic:
				m_sizeGrafic += GetData(data, &m_memGrafic[address]);
				break;

			case TypeEOF:
				break;
			}
		}
	}
	fclose(fp);	

	// Member durch gelesene Daten initialisieren und pruefen
	if (m_sizeGrafic) {
		
		int i, j;

		// 2.
		// Grafiken pruefen

		// Anzahl der Grafiken setzen
		i = m_pGraficNode->cnt; 

		if (i > MAX_GRAFIC) {

			setErrorState( Format );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
			return 0;
		}

		// Zeigerarray und Grafikstrukturzeiger initialisieren
		for (j=0; j < i; j++) {
			
			m_pGrafic[j] = (GRAFIC*)&m_memGrafic[m_pGraficNode->ptrGraficArray + i * sizeof(unsigned int) + j * sizeof(GRAFIC)];
		}

		// CRC32 pruefen
		unsigned char ucByteCount;
		for (j=0; j < i; j++) {
	
			ucByteCount = (unsigned char)((m_pGrafic[j]->width - 1) / 8) + 1;

			if (m_pGrafic[j]->crc != GetCRC32(&m_memGrafic[m_pGrafic[j]->ptrGrafic], ucByteCount * m_pGrafic[j]->height)) {

				setErrorState( CrcFail );
				setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
				return 0;
			}
		}

	} else {

		setErrorState( NoData );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
		return 0;
	}

	return 1;
}

int FileDisplay::Save( const std::string &filePath, bool append ) 
{
	FILE *fp;
	char szData[0x400];

	// Fehlerobjekt zuruecksetzen
	resetError();

	// vorhandene Datei anlegen
	if (!append) {
		if ((fp = fopen(filePath.data(), "w")) == (FILE*)0) {
			// Oeffnen der Datei gescheitert
			setErrorState( FileOpen );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
			return 0;
		}
		fclose(fp);
	}

	// Daten der Zeichensaetze sichern
	if (!SaveFont(filePath, true)) {

		return 0;
	}

	// Daten der Grafiken sichern
	if (!SaveGrafic(filePath, true)) {

		return 0;
	}

	if (!append) {
		// Endezeile in Datei eintragen
		if ((fp = fopen(filePath.data(), "a")) == (FILE*)0) {
			// Oeffnen der Datei gescheitert
			setErrorState( FileOpen );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
			return 0;
		}

		if (!MakeEndLine(szData)) {
			
			fclose(fp);
			setErrorState( RecordEndMake );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
			return 0;
		}

		fprintf(fp, "%s\n", szData);
		fclose(fp);
	}

	return 1;
}

int FileDisplay::SaveFont(const std::string &filePath, bool append) 
{
	FILE *fp;
	char szData[0x400];

	if ((fp = fopen(filePath.data(), (append) ? "a" : "w")) == (FILE*)0) {
		// Oeffnen der Datei gescheitert
		setErrorState( FileOpen );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
		return 0;
	}

	// Daten der Zeichensaetze eintragen
	unsigned int ulPos;
	for (ulPos=0; ulPos < m_sizeFont; ulPos += maxDataSize()) {

		if (!MakeTypedDataLine((unsigned short)ulPos, TypeDisplayFont, 
			&m_memFont[ulPos], (unsigned char)(((ulPos + maxDataSize()) < m_sizeFont) ? maxDataSize() : m_sizeFont - ulPos), szData)) {
			
			setErrorState( RecordMake );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
			break;
		}

		fprintf(fp, "%s\n", szData);
	}

	// Abbruch nach Fehler
	if (errorState() != None) {

		fclose(fp);
		return 0;
	}

	if (!append) {

		if (!MakeEndLine(szData)) {
			
			fclose(fp);
			setErrorState( RecordEndMake );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
			return 0;
		}

		fprintf(fp, "%s\n", szData);
	}

	fclose(fp);

	char fName[260];
	sprintf(fName, "%s.font.bin", filePath.data());

	// Datei als Binär rausschreiben
	if ((fp = fopen(fName, "wb")) == (FILE*)0) {
		// Oeffnen der Datei gescheitert
		setErrorState( FileOpen );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
		return 0;
	}

	fwrite(m_memFont, 1, m_sizeFont, fp);

	fclose(fp);

	return 1;
}

int FileDisplay::SaveGrafic( const std::string &filePath, bool append ) 
{
	FILE *fp;
	char szData[0x400];

	if ((fp = fopen(filePath.data(), (append) ? "a" : "w")) == (FILE*)0) {
		// Oeffnen der Datei gescheitert
		setErrorState( FileOpen );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
		return 0;
	}

	// Daten der Bilder eintragen
	unsigned int ulPos;
	for (ulPos=0; ulPos < m_sizeGrafic; ulPos += maxDataSize()) {

		if (!MakeTypedDataLine((unsigned short)ulPos, TypeDisplayGrafic, 
			&m_memGrafic[ulPos], (unsigned char)(((ulPos + maxDataSize()) < m_sizeGrafic) ? maxDataSize() : m_sizeGrafic - ulPos), szData)) {
			
			setErrorState( RecordMake );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
			break;
		}

		fprintf(fp, "%s\n", szData);
	}

	// Abbruch nach Fehler
	if (errorState() != None) {

		fclose(fp);
		return 0;
	}

	if (!append) {

		if (!MakeEndLine(szData)) {
			
			fclose(fp);
			setErrorState( RecordEndMake );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
			return 0;
		}

		fprintf(fp, "%s\n", szData);
	}

	fclose(fp);

    char fName[260];
    sprintf(fName, "%s.grafic.bin", filePath.data());

    // Datei als Binär rausschreiben
    if ((fp = fopen(fName, "wb")) == (FILE*)0) {
        // Oeffnen der Datei gescheitert
		setErrorState( FileOpen );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return 0;
    }

    fwrite(m_memGrafic, 1, m_sizeGrafic, fp);

    fclose(fp);

	return 1;
}

// Textdaten als letzten Zeichensatz anfuegen.
int FileDisplay::AppendSegmentText(unsigned char *buf, unsigned int size) 
{
	(void)size;
	unsigned char *memFont;
	// Verwaltungsstrukturen fuer Zeichensatz
	FONTNODE *pFontNode;
	FONT     *pFont[MAX_FONT];
	TEXTDATA *pText = (TEXTDATA*)buf; // Ist TEXTDATA hat aber den gleichen Aufbau wie FONT
	unsigned int memFontSize;
	int i = 0;

	// Sind bereits Zeichensaetze angegeben?
	if (m_sizeFont) {

		//###################################
		// Als letzten Zeichensatz anfuegen

		// Speichergroesse von Zeichensaetzen berechnen
		memFontSize = m_pFont[m_pFontNode->cnt - 1]->ptrCharacter + m_pFont[m_pFontNode->cnt - 1]->size;
		// Kopie der Bereits vorhandenen Zeichensatzdaten erstellen
		memFont = new unsigned char[memFontSize];
		// Speicher allokation nicht erfolgreich?
		if (memFont == NULL) {
			setErrorState( AllocMem );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
			return 0;
		}

		// Speicher kopieren
		memcpy(memFont, m_memFont, memFontSize);

		pFontNode = (FONTNODE*)memFont;
		// Kontrollstrukturen initialisieren
		for (i=0; i < pFontNode->cnt; i++) {

			pFont[i] = (FONT*)&memFont[pFontNode->ptrFontArray + pFontNode->cnt*sizeof(unsigned int) + i * sizeof(FONT)];
		}

		// Zeichensaetze initialisieren
		DefaultFont(pFontNode->cnt + 1);

		// Strukturen der Bestehende Zeichensaetze uebernehmen
		for (i=0; i < pFontNode->cnt; i++) {

			*m_pFont[i] = *pFont[i];
		}
		// Neuen Textblock anfuegen
		*m_pFont[m_pFontNode->cnt - 1] = *(FONT*)buf;

		// Erstes Freies Datenbyte
		m_pFont[0]->ptrTable = m_pFontNode->ptrFontArray + m_pFontNode->cnt*sizeof(unsigned int) + m_pFontNode->cnt*sizeof(FONT);
		m_pFont[0]->ptrCharacter = m_pFont[0]->ptrTable + 0x100;;

		// Daten der bestehenden Zeichensaetze uebernehmen
		for (i=0; i < pFontNode->cnt; i++) {

			memcpy(&m_memFont[m_pFont[i]->ptrTable], &memFont[pFont[i]->ptrTable], 0x100);
			memcpy(&m_memFont[m_pFont[i]->ptrCharacter], &memFont[pFont[i]->ptrCharacter], pFont[i]->size);
			
			// Naechsten freien Speicher zuweisen
			m_pFont[i + 1]->ptrTable = m_pFont[i]->ptrCharacter + m_pFont[i]->size;
			m_pFont[i + 1]->ptrCharacter = m_pFont[i + 1]->ptrTable + 0x100;;
		}

		// Neuen Block anfuegen
		memcpy(&m_memFont[m_pFont[i]->ptrTable], &buf[pText->ptrIDTable], pText->cnt*sizeof(unsigned short));
		m_pFont[i]->ptrCharacter = m_pFont[i]->ptrTable + pText->cnt*sizeof(unsigned short);
		memcpy(&m_memFont[m_pFont[i]->ptrCharacter], &buf[pText->ptrTextdata], pText->size);

		// Speicher wieder freigeben
		delete[] memFont;

		// Groessen anpassen
		m_sizeFont = m_pFont[m_pFontNode->cnt - 1]->ptrCharacter + m_pFont[m_pFontNode->cnt - 1]->size;
		m_sizeData = m_sizeFont + m_sizeGrafic;

	} else {

		//###################################
		// Als einzelnen Zeichensatz anfuegen

		// Zeichensaetze initialisieren
		DefaultFont(1);

		// Neuen Textblock anfuegen
		*m_pFont[0] = *(FONT*)buf;

		// Erstes Freies Datenbyte
		m_pFont[0]->ptrTable = m_pFontNode->ptrFontArray + m_pFontNode->cnt*sizeof(unsigned int) + m_pFontNode->cnt*sizeof(FONT);
		m_pFont[0]->ptrCharacter = m_pFont[0]->ptrTable + pText->cnt*sizeof(unsigned short);

		// Neuen Block anfuegen
		memcpy(&m_memFont[m_pFont[0]->ptrTable], &buf[pText->ptrIDTable], pText->cnt*sizeof(unsigned short));
		memcpy(&m_memFont[m_pFont[0]->ptrCharacter], &buf[pText->ptrTextdata], pText->size);

		// Groessen anpassen
		m_sizeFont = m_pFont[0]->ptrCharacter + m_pFont[0]->size;
		m_sizeData = m_sizeFont + m_sizeGrafic;
	}

	// BUGFIX: Anpassung der Groesse, damit die CRC-Berechnung von der Firmware
	// durchgefuehrt werden kann. Die Firmware geht von einer festen Tabellengroesse von 256 Byte aus.
	// Bei den Textdaten kann die Groesse dieser Tabelle abhaengig von der Maximal-ID variieren.
	m_pFont[i]->size += (((TEXTDATA*)m_pFont[i])->cnt * sizeof(unsigned short)) - 256;

	return 1;
}

